import os
from typing import Dict, List, Optional, Set

from pydantic.types import FilePath
from pydantic import BaseModel, DirectoryPath
from enum import Enum
from pathlib import PosixPath, WindowsPath

from core.utils import paths as core_paths, hasher as cdev_hasher
from core.utils.logger import log
from core.utils.platforms import lambda_python_environment

import sys
from parsley import makeGrammar

from rich import print



def get_lines_from_file_list(file_list, function_info) -> List[str]:
    # Get the list of lines from a file based on the function info provided
    line_nos = _compress_lines(function_info)

    actual_lines = []

    for i in line_nos:
        if i == -1:
            actual_lines.append(os.linesep)
        elif i <= len(file_list):
            actual_lines.append(file_list[i - 1])

    return actual_lines


def get_file_as_list(path):
    # Returns the file as a list of lines
    if not os.path.isfile:
        return None

    with open(path) as fh:
        rv = fh.read().splitlines()

    return rv


def _compress_lines(original_lines):
    # Takes input SORTED([(l1,l2), (l3,l4), ...])
    # returns [l1,...,l2,l3,...,l4]
    rv = []

    for pair in original_lines:
        for i in range(pair[0], pair[1] + 1):
            if rv and rv[-1] == i:
                # if the last element already equals the current value continue... helps eleminate touching boundaries
                continue

            rv.append(i)

        if sys.version_info > (3, 8):
            rv.append(-1)

    return rv


def get_parsed_path(original_path, function_name, final_base_directory: DirectoryPath, prefix=None):
    split_path = core_paths.get_relative_to_workspace_path(original_path).split("/")
    
    # the last item in the path is .py file name... change the  .py to _py so it works as a dir
    final_file_name = split_path[-1].split(".")[0] + "_" + function_name + ".py"
    
    try:
        split_path.remove(".")
        split_path.remove("..")
    except Exception as e:
        pass

    if prefix:
        split_path.insert(0, prefix)

    final_file_dir = core_paths.create_path(final_base_directory, split_path[:-1])


    log.debug("Created Parsed Path is %s", os.path.join(final_file_dir, final_file_name) )
    return os.path.join(final_file_dir, final_file_name)


class ExternalDependencyWriteInfo(BaseModel):
    location: str
    id: str

    class Config:
        json_encoders = {
            PosixPath: lambda v: v.as_posix(),  # or lambda v: str(v)
            WindowsPath: lambda v: v.as_posix(),
        }

        extra = "ignore"

    def __hash__(self) -> int:
        return int(cdev_hasher.hash_string(self.id), base=16)


class PackageTypes(str, Enum):
    BUILTIN = "builtin"
    STANDARDLIB = "standardlib"
    PIP = "pip"
    LOCALPACKAGE = "localpackage"
    AWSINCLUDED = "awsincluded"

class Architecture(str, Enum):
    X86 = "x86_64"
    ARM64 = "aarch64"
    ANY = "any"



class ModulePackagingInfo(BaseModel):
    module_name: str
    type: PackageTypes
    version_id: Optional[str]
    arch: Optional[Architecture] = Architecture.ANY
    fp: Optional[str]
    tree: Optional[List["ModulePackagingInfo"]]
    flat: Optional[List["ModulePackagingInfo"]]
    is_relative: Optional[bool]

    def set_flat(self, flat: List["ModulePackagingInfo"]):
        self.flat = flat

    def set_tree(self, tree: List["ModulePackagingInfo"]):
        self.tree = tree

    def get_id_str(self) -> str:
        if self.type == PackageTypes.LOCALPACKAGE:
            if os.path.isfile(self.fp):
                return f"{self.module_name}-{self.fp}-{cdev_hasher.hash_file(self.fp)}"

            else:
                return f"{self.module_name}-{self.fp}"

        elif self.type == PackageTypes.PIP:
            return f"{self.module_name}-{self.version_id}-{self.arch}"

        else:
            return self.module_name

    def __str__(self) -> str:
        return self.get_id_str()

    def __repr__(self) -> str:
        return f"{self.get_id_str()}"

    def __hash__(self) -> int:
        return int(cdev_hasher.hash_string(self.get_id_str()), base=16)


ModulePackagingInfo.update_forward_refs()


_depth_to_color = {0: "white", 1: "blue", 2: "yellow", 3: "magenta", 4: "red"}


def print_dependency_tree(
    handler_name: str, top_level_modules: List[ModulePackagingInfo]
):
    print(f"Handler: {handler_name}")

    for module_info in top_level_modules:

        _recursive_dfs_print(module_info, 0)
        print("|")


def _recursive_dfs_print(module_info: ModulePackagingInfo, depth: int):

    base_str = f"|[{_depth_to_color.get(depth%5)}]{'-' * depth } {module_info.get_id_str()}[/{_depth_to_color.get(depth%5)}] ({module_info.type})"
    print(base_str)

    if module_info.tree:
        for child_module in module_info.tree:
            _recursive_dfs_print(child_module, depth + 1)


def print_handler_package():
    pass


def print_layer_package():
    pass




environment_to_architecture_suffix: Dict[lambda_python_environment, Architecture] = {
    lambda_python_environment.py37: Architecture.X86,
    lambda_python_environment.py38_x86_64: Architecture.X86,
    lambda_python_environment.py38_arm64: Architecture.ARM64,
    lambda_python_environment.py39_x86_64: Architecture.X86,
    lambda_python_environment.py39_arm64: Architecture.ARM64,
    lambda_python_environment.py3_x86_64: Architecture.X86,
    lambda_python_environment.py3_arm64: Architecture.ARM64,
}


CONTAINER_NAMES = {
    lambda_python_environment.py37: "public.ecr.aws/lambda/python:3.7",
    lambda_python_environment.py38_x86_64: "public.ecr.aws/lambda/python:3.8-x86_64",
    lambda_python_environment.py38_arm64: "public.ecr.aws/lambda/python:3.8-arm64",
    lambda_python_environment.py39_x86_64: "public.ecr.aws/lambda/python:3.9-x86_64",
    lambda_python_environment.py39_arm64: "public.ecr.aws/lambda/python:3.9-arm64",
    lambda_python_environment.py3_x86_64: "public.ecr.aws/lambda/python:3-x86_64",
    lambda_python_environment.py3_arm64: "public.ecr.aws/lambda/python:3-arm64",
}


# https://www.python.org/dev/peps/pep-0508/
grammar = """
    wsp           = ' ' | '\t'
    version_cmp   = wsp* <'<=' | '<' | '!=' | '==' | '>=' | '>' | '~=' | '==='>
    version       = wsp* <( letterOrDigit | '-' | '_' | '.' | '*' | '+' | '!' )+>
    version_one   = version_cmp:op version:v wsp* -> (op, v)
    version_many  = version_one:v1 (wsp* ',' version_one)*:v2 -> [v1] + v2
    versionspec   = ('(' version_many:v ')' ->v) | version_many
    urlspec       = '@' wsp* <URI_reference>
    marker_op     = version_cmp | (wsp* 'in') | (wsp* 'not' wsp+ 'in') -> 'not in'
    python_str_c  = (wsp | letter | digit | '(' | ')' | '.' | '{' | '}' |
                     '-' | '_' | '*' | '#' | ':' | ';' | ',' | '/' | '?' |
                     '[' | ']' | '!' | '~' | '`' | '@' | '$' | '%' | '^' |
                     '&' | '=' | '+' | '|' | '<' | '>' )
    dquote        = '"'
    squote        = '\\''
    python_str    = (squote <(python_str_c | dquote)*>:s squote |
                     dquote <(python_str_c | squote)*>:s dquote) -> s
    env_var       = ('python_version' | 'python_full_version' |
                     'os_name' | 'sys_platform' | 'platform_release' |
                     'platform_system' | 'platform_version' |
                     'platform_machine' | 'platform_python_implementation' |
                     'implementation_name' | 'implementation_version' |
                     'extra' # ONLY when defined by a containing layer
                     ):varname -> lookup(varname)
    marker_var    = wsp* (env_var | python_str)
    marker_expr   = marker_var:l marker_op:o marker_var:r -> (o, l, r)
                  | wsp* '(' marker:m wsp* ')' -> m
    marker_and    = marker_expr:l wsp* 'and' marker_expr:r -> ('and', l, r)
                  | marker_expr:m -> m
    marker_or     = marker_and:l wsp* 'or' marker_and:r -> ('or', l, r)
                      | marker_and:m -> m
    marker        = marker_or
    quoted_marker = ';' wsp* marker
    identifier_end = letterOrDigit | (('-' | '_' | '.' )* letterOrDigit)
    identifier    = < letterOrDigit identifier_end* >
    name          = identifier
    extras_list   = identifier:i (wsp* ',' wsp* identifier)*:ids -> [i] + ids
    extras        = '[' wsp* extras_list?:e wsp* ']' -> e
    name_req      = (name:n wsp* extras?:e wsp* versionspec?:v wsp* quoted_marker?:m
                     -> (n, e or [], v or [], m))
    url_req       = (name:n wsp* extras?:e wsp* urlspec:v (wsp+ | end) quoted_marker?:m
                     -> (n, e or [], v, m))
    specification = wsp* ( url_req | name_req ):s wsp* -> s
    # The result is a tuple - name, list-of-extras,
    # list-of-version-constraints-or-a-url, marker-ast or None


    URI_reference = <URI | relative_ref>
    URI           = scheme ':' hier_part ('?' query )? ( '#' fragment)?
    hier_part     = ('//' authority path_abempty) | path_absolute | path_rootless | path_empty
    absolute_URI  = scheme ':' hier_part ( '?' query )?
    relative_ref  = relative_part ( '?' query )? ( '#' fragment )?
    relative_part = '//' authority path_abempty | path_absolute | path_noscheme | path_empty
    scheme        = letter ( letter | digit | '+' | '-' | '.')*
    authority     = ( userinfo '@' )? host ( ':' port )?
    userinfo      = ( unreserved | pct_encoded | sub_delims | ':')*
    host          = IP_literal | IPv4address | reg_name
    port          = digit*
    IP_literal    = '[' ( IPv6address | IPvFuture) ']'
    IPvFuture     = 'v' hexdig+ '.' ( unreserved | sub_delims | ':')+
    IPv6address   = (
                      ( h16 ':'){6} ls32
                      | '::' ( h16 ':'){5} ls32
                      | ( h16 )?  '::' ( h16 ':'){4} ls32
                      | ( ( h16 ':')? h16 )? '::' ( h16 ':'){3} ls32
                      | ( ( h16 ':'){0,2} h16 )? '::' ( h16 ':'){2} ls32
                      | ( ( h16 ':'){0,3} h16 )? '::' h16 ':' ls32
                      | ( ( h16 ':'){0,4} h16 )? '::' ls32
                      | ( ( h16 ':'){0,5} h16 )? '::' h16
                      | ( ( h16 ':'){0,6} h16 )? '::' )
    h16           = hexdig{1,4}
    ls32          = ( h16 ':' h16) | IPv4address
    IPv4address   = dec_octet '.' dec_octet '.' dec_octet '.' dec_octet
    nz            = ~'0' digit
    dec_octet     = (
                      digit # 0-9
                      | nz digit # 10-99
                      | '1' digit{2} # 100-199
                      | '2' ('0' | '1' | '2' | '3' | '4') digit # 200-249
                      | '25' ('0' | '1' | '2' | '3' | '4' | '5') )# %250-255
    reg_name = ( unreserved | pct_encoded | sub_delims)*
    path = (
            path_abempty # begins with '/' or is empty
            | path_absolute # begins with '/' but not '//'
            | path_noscheme # begins with a non-colon segment
            | path_rootless # begins with a segment
            | path_empty ) # zero characters
    path_abempty  = ( '/' segment)*
    path_absolute = '/' ( segment_nz ( '/' segment)* )?
    path_noscheme = segment_nz_nc ( '/' segment)*
    path_rootless = segment_nz ( '/' segment)*
    path_empty    = pchar{0}
    segment       = pchar*
    segment_nz    = pchar+
    segment_nz_nc = ( unreserved | pct_encoded | sub_delims | '@')+
                    # non-zero-length segment without any colon ':'
    pchar         = unreserved | pct_encoded | sub_delims | ':' | '@'
    query         = ( pchar | '/' | '?')*
    fragment      = ( pchar | '/' | '?')*
    pct_encoded   = '%' hexdig
    unreserved    = letter | digit | '-' | '.' | '_' | '~'
    reserved      = gen_delims | sub_delims
    gen_delims    = ':' | '/' | '?' | '#' | '(' | ')?' | '@'
    sub_delims    = '!' | '$' | '&' | '\\'' | '(' | ')' | '*' | '+' | ',' | ';' | '='
    hexdig        = digit | 'a' | 'A' | 'b' | 'B' | 'c' | 'C' | 'd' | 'D' | 'e' | 'E' | 'f' | 'F'
"""

bindings = {
    "py6_86": {
        "implementation_name": "cpython",
        "implementation_version": "3.6.15",
        "os_name": "posix",
        "platform_machine": "x86_64",
        "platform_python_implementation": "CPython",
        "platform_release": "4.14.246-198.474.amzn2.x86_64",
        "platform_system": "Linux",
        "platform_version": "#1 SMP Wed Sep 15 00:16:00 UTC 2021",
        "python_full_version": "3.6.15",
        "python_version": "3.6",
        "sys_platform": "linux",
    },
    "py7_86": {
        "implementation_name": "cpython",
        "implementation_version": "3.7.12",
        "os_name": "posix",
        "platform_machine": "x86_64",
        "platform_python_implementation": "CPython",
        "platform_release": "4.14.246-198.474.amzn2.x86_64",
        "platform_system": "Linux",
        "platform_version": "#1 SMP Wed Sep 15 00:16:00 UTC 2021",
        "python_full_version": "3.7.12",
        "python_version": "3.7",
        "sys_platform": "linux",
    },
    "py8_arm": {
        "implementation_name": "cpython",
        "implementation_version": "3.8.11",
        "os_name": "posix",
        "platform_machine": "aarch64",
        "platform_python_implementation": "CPython",
        "platform_release": "4.14.246-198.474.amzn2.aarch64",
        "platform_system": "Linux",
        "platform_version": "#1 SMP Wed Sep 15 00:16:11 UTC 2021",
        "python_full_version": "3.8.11",
        "python_version": "3.8",
        "sys_platform": "linux",
    },
    "py8_86": {
        "implementation_name": "cpython",
        "implementation_version": "3.8.11",
        "os_name": "posix",
        "platform_machine": "x86_64",
        "platform_python_implementation": "CPython",
        "platform_release": "4.14.246-198.474.amzn2.x86_64",
        "platform_system": "Linux",
        "platform_version": "#1 SMP Wed Sep 15 00:16:00 UTC 2021",
        "python_full_version": "3.8.11",
        "python_version": "3.8",
        "sys_platform": "linux",
    },
    "py9_arm": {
        "implementation_name": "cpython",
        "implementation_version": "3.9.6",
        "os_name": "posix",
        "platform_machine": "aarch64",
        "platform_python_implementation": "CPython",
        "platform_release": "4.14.246-198.474.amzn2.aarch64",
        "platform_system": "Linux",
        "platform_version": "#1 SMP Wed Sep 15 00:16:11 UTC 2021",
        "python_full_version": "3.9.6",
        "python_version": "3.9",
        "sys_platform": "linux",
    },
    "py9_86": {
        "implementation_name": "cpython",
        "implementation_version": "3.9.6",
        "os_name": "posix",
        "platform_machine": "x86_64",
        "platform_python_implementation": "CPython",
        "platform_release": "4.14.246-198.474.amzn2.x86_64",
        "platform_system": "Linux",
        "platform_version": "#1 SMP Wed Sep 15 00:16:00 UTC 2021",
        "python_full_version": "3.9.6",
        "python_version": "3.9",
        "sys_platform": "linux",
    },
}

compiled = makeGrammar(grammar, {"lookup": bindings.get("py8_arm").__getitem__})


def parse_requirement_line(line: str) -> str:
    """
    From a dist-require line parse out the name of the need project.
    """
    # All this code is directly lifted from PEP 508 which is the specification of this line.
    # https://www.python.org/dev/peps/pep-0508/
    try:
        parsed_info = compiled(line).specification()
        
        if parsed_info[3]:
            if not evaluate_expression(parsed_info[3]):
                return None

        return parsed_info[0]

    except Exception:
        return None


OPERATORS = set(
    [
        "in",
        "==",
        "!=",
        "not in",
        "and",
        "or",
        "<=",
        "<",
        "!=",
        "==",
        ">=",
        ">",
    ]
)


def evaluate_expression(expr) -> bool:
    operator = expr[0]
    right_side = expr[1]
    left_side = expr[2]

    if operator not in OPERATORS:
        raise Exception

    if operator == "and":
        right_side_evaluated = evaluate_expression(right_side)

        if not right_side_evaluated:
            return False

        left_side_evaluated = evaluate_expression(left_side)

        return right_side_evaluated and left_side_evaluated

    elif operator == "or":
        right_side_evaluated = evaluate_expression(right_side)

        if right_side_evaluated:
            return True

        left_side_evaluated = evaluate_expression(left_side)

        return right_side_evaluated or left_side_evaluated

    elif operator == "in":
        return right_side in left_side

    elif operator == "==":
        return right_side == left_side

    elif operator == "!=":
        return right_side != left_side

    elif operator == "not in":
        return not right_side in left_side

    elif operator == "<=":
        return right_side <= left_side

    elif operator == "<":
        return right_side < left_side

    elif operator == "!=":
        return right_side != left_side

    elif operator == "==":
        return right_side == left_side

    elif operator == ">=":
        return right_side >= left_side

    elif operator == ">":
        return right_side > left_side

    else:
        raise Exception



"""Utilities to provide a 'standardized' way of producing hashes to identify uniqueness. 

These hash functions are **not**
cryptographically secure and should not be used for that. Instead, these functions should simply be used to create hashes
that will be used to track changes throughout the system. 
"""

import hashlib
import os
from typing import List

from pydantic.types import FilePath

from core.utils.exceptions import Cdev_Error


class FILE_CACHE_CLASS:
    cache = {}


FILE_CACHE = FILE_CACHE_CLASS()


def hash_list(val: List[str], deliminator: str = ";") -> str:
    """Hash a list of str values

    Note that the order of the list also determines the output, therefore if the input is stored in
    an unsorted collection, you should used `hash_set` instead.

    Args:
        val (List[str]): The list of str to produce a hash of
        deliminator (Optional[str]): A value use to seperate the input. Defaults to ';'

    Returns:
        str: hash of the values
    """

    return hash_string(deliminator.join([str(x) for x in val]))


def hash_string(val: str) -> str:
    """Hash a str value

    Implemented using md5.

    Args:
        val (str): value to hash

    Returns:
        str: hash of the value
    """
    return hashlib.md5(val.encode()).hexdigest()


def clear_file_cache():
    """Clear the cache used by the `hash_file` utility."""
    FILE_CACHE.cache = {}


def hash_file(fp: FilePath, bypass_cache: bool = False) -> str:
    """Hash a file given a path

    Note that the implementation contains a reference to a cache. Since this utility is primarily 
    used as an internal tool with the framework, the cache helps speeds things up when we know the file has
    not changed. Note that the framework periodically flushes this cache when needed within the context of 
    the execution of the framework using the `clear_file_cache` function. 

    If using this outside the confides of the framework, you can by pass the cache by setting the `bypass_cache`
    flag.

    Note that the implementation returns a md5 hash of the bytes in the file. 

    Args:
        fp (FilePath): Path to the file. Must be a resolvable path on the filesystem.
        bypass_cache (Optional[bool]): By pass the internal cache. Default False.

    Returns:
        str: hash of the file

    Raises:
        Cdev_Error
    """
    if fp in FILE_CACHE.cache and not bypass_cache:
        return FILE_CACHE.cache.get(fp)


    if not os.path.isfile(fp):
        raise Cdev_Error(f"Could not find file ({fp}) to hash", FileNotFoundError)

    with open(fp, "rb") as fh:
        hash = hashlib.md5(fh.read()).hexdigest()

    FILE_CACHE.cache[fp] = hash

    return hash



"""


This tool is designed to improve the developer experience for Serverless development. 


[![Demo](https://img.shields.io/badge/demo-holder-blue)](https://img.shields.io/badge/demo-holder-blue)

"""


# from . import commands, cli, settings, utils, constructs, models
#
#
# __all__ = [ "cli", "settings", "commands", "utils", "constructs", "models"]

__pdoc__ = {}
__pdoc__[".venv"] = False


from . import constructs, resources


# Ergonomic mapping so that the global project instance is in a more logical place for end developers.
Project = constructs.project.Project




from enum import Enum
from typing import Any, Dict, List
from uuid import uuid4

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.default.resources.simple import object_store as simple_object_store
from core.constructs.output_manager import OutputTask
from core.utils import hasher


from .. import aws_client as raw_aws_client


RUUID = "cdev::simple::bucket"


class event_hander_types(Enum):
    SQS = "sqs"
    LAMBDA = "lambda"
    SNS = "sns"


def _create_simple_bucket(
    transaction_token: str, 
    namespace_token: str, 
    resource:  simple_object_store.bucket_model,
    output_task: OutputTask
) -> Dict:

    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])

    cloud_bucket_name = f"cdev-bucket-{full_namespace_suffix}"

    output_task.update(comment=f'Creating Bucket {cloud_bucket_name}')
    
    # TODO create buckets in different region
    rv = raw_aws_client.run_client_function(
        "s3", "create_bucket", {"Bucket": cloud_bucket_name}
    )

    output_task.update(comment=f'Created Bucket {cloud_bucket_name}')

    output_info = {
        "bucket_name": cloud_bucket_name,
        "cloud_id": f"arn:aws:s3:::{cloud_bucket_name}",
        "arn": f"arn:aws:s3:::{cloud_bucket_name}",
    }


    return output_info


def _update_simple_bucket(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: simple_object_store.bucket_model,
    new_resource: simple_object_store.bucket_model,
    previous_output: Dict,
    output_task: OutputTask
) -> Dict:

    _remove_simple_bucket(transaction_token,  previous_output, output_task)
    new_info =  _create_simple_bucket(transaction_token, namespace_token, new_resource, output_task)

    return new_info


def _remove_simple_bucket(
    transaction_token: str,  previous_output: Dict, output_task: OutputTask
):

    previous_bucket_name = previous_output.get('bucket_name')

    output_task.update(advance=1, comment=f'Deleting Bucket {previous_bucket_name}')

    raw_aws_client.run_client_function(
        "s3", "delete_bucket", {"Bucket": previous_bucket_name}
    )

    output_task.update(advance=1, comment=f'Deleted Bucket {previous_bucket_name}')


def add_eventsource(
    bucket_name: str,
    handler_type: event_hander_types,
    handler_arn: str,
    events: List[str],
) -> str:
    """Add a new Event Source to a bucket


    Per https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_bucket_notification_configuration
    The only available function for adding event handlers to a bucket is 'put_bucket_notification_configuration", but this function takes
    in a whole configuration and completely overwrites to the new one. Therefor, we need to get the current configuration and then add this
    new event handler and write it back to preserve previous eventsources.
    """

    current_events = _get_current_bucket_event_configuration(bucket_name)

    if not current_events:
        current_events = {}


    event_id = hasher.hash_list(
        [handler_type.value, handler_arn, hasher.hash_list(events)]
    )
    base_config = {"Events": events, "Id": event_id}

    if handler_type == event_hander_types.LAMBDA:
        base_config.update({"LambdaFunctionArn": handler_arn})

        current_lambda_handlers = current_events.get("LambdaFunctionConfigurations")

        if not current_lambda_handlers:
            current_lambda_handlers = []

        current_lambda_handlers.append(base_config)
        current_events["LambdaFunctionConfigurations"] = current_lambda_handlers
    elif handler_type == event_hander_types.SNS:
        base_config.update({"TopicArn": handler_arn})

        current_sns_handlers = current_events.get("TopicConfigurations")

        if not current_sns_handlers:
            current_sns_handlers = []

        current_sns_handlers.append(base_config)
        current_events["TopicConfigurations"] = current_sns_handlers
    elif handler_type == event_hander_types.SQS:
        base_config.update({"QueueArn": handler_arn})

        current_sqs_handlers = current_events.get("QueueConfigurations")

        if not current_sqs_handlers:
            current_sqs_handlers = []

        current_sqs_handlers.append(base_config)
        current_events["QueueConfigurations"] = current_sqs_handlers
    else:
        raise Exception

    _deploy_bucket_event_configuration(bucket_name, current_events)

    return event_id


def remove_eventsource(bucket_name: str, bucket_event_id: str) -> bool:
    current_events = _get_current_bucket_event_configuration(bucket_name)

    if not current_events:
        raise Exception


    keys = [
        "TopicConfigurations",
        "QueueConfigurations",
        "LambdaFunctionConfigurations",
    ]

    found_val = False
    for key in keys:
        if found_val:
            break

        if not current_events.get(key):
            continue

        for val in current_events.get(key):
            if val.get("Id") == bucket_event_id:
                remove_val = val
                remove_key = key
                found_val = True
                break

    if not found_val:
        raise Exception

    current_events.get(remove_key).remove(remove_val)

    _deploy_bucket_event_configuration(bucket_name, current_events)

    return True


def _get_current_bucket_event_configuration(bucket_name: str) -> Dict:
    aws_rv = raw_aws_client.run_client_function(
        "s3", "get_bucket_notification_configuration", {"Bucket": bucket_name}
    )

    rv = {
        "TopicConfigurations": aws_rv.get("TopicConfigurations"),
        "QueueConfigurations": aws_rv.get("QueueConfigurations"),
        "LambdaFunctionConfigurations": aws_rv.get("LambdaFunctionConfigurations"),
    }

    return rv


def _deploy_bucket_event_configuration(bucket_name: str, new_config: Dict):
    final_config = {k: v for k, v in new_config.items() if v}

    raw_aws_client.run_client_function(
        "s3",
        "put_bucket_notification_configuration",
        {"Bucket": bucket_name, "NotificationConfiguration": final_config},
    )



def handle_simple_bucket_deployment( 
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_bucket(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
        return _update_simple_bucket(
            transaction_token,
            namespace_token,
            simple_object_store.bucket_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )

    elif resource_diff.action_type == Resource_Change_Type.DELETE:
        _remove_simple_bucket(
            transaction_token,
            previous_output,
            output_task
        )

        return {}

    



from typing import Any, Dict, List
from uuid import uuid4

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.default.resources.simple import queue 
from core.constructs.output_manager import OutputTask
from core.utils import hasher

from .. import aws_client

def _create_simple_queue(
    transaction_token: str, 
    namespace_token: str, 
    resource:  queue.queue_model,
    output_task: OutputTask
) -> Dict:

    attributes = {}

    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])

    queue_name = f"cdev-queue-{full_namespace_suffix}.fifo" if resource.is_fifo else f"cdev-queue-{full_namespace_suffix}"

    if resource.is_fifo:
        attributes.update({"FifoQueue": str(resource.is_fifo)})

    output_task.update(comment=f'Creating Queue {queue_name}')

    rv = aws_client.run_client_function(
        "sqs",
        "create_queue",
        {
            "QueueName": queue_name, 
            "Attributes": attributes
        },
    )

   
    output_info = {
        "queue_name": queue_name,
        "cdev_name": resource.name,
        "queue_url": rv.get("QueueUrl"),
    }

    # rv does not contain the arn of the queue so we need a second call

    rv = aws_client.run_client_function(
        "sqs",
        "get_queue_attributes",
        {"QueueUrl": output_info.get("queue_url"), "AttributeNames": ["QueueArn"]},
    )

    output_info.update(
        {
            "cloud_id": rv.get("Attributes").get("QueueArn"),
            "arn": rv.get("Attributes").get("QueueArn"),
        }
    )

    return output_info


def _update_simple_queue(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: queue.queue_model,
    new_resource: queue.queue_model,
    previous_output: Dict,
    output_task: OutputTask
) -> Dict:
    _remove_simple_queue(transaction_token, previous_output, output_task)
    new_info = _create_simple_queue(transaction_token, namespace_token, new_resource, output_task)
    return new_info


def _remove_simple_queue(
    transaction_token: str, previous_output: Dict, output_task: OutputTask
):
    queue_url = previous_output.get("queue_url")

    output_task.update(comment=f'Deleting Queue')

    aws_client.run_client_function(
        "sqs", 
        "delete_queue",
        {"QueueUrl": queue_url}
    )


def handle_simple_queue_deployment(
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_queue(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
        return _update_simple_queue(
            transaction_token,
            namespace_token,
            queue.queue_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )

    elif resource_diff.action_type == Resource_Change_Type.DELETE:
        _remove_simple_queue(
            transaction_token,
            previous_output,
            output_task
        )

        return {}



"""Construct that represents the collection of other primitives 


"""

from enum import Enum
import inspect
from rich.console import Console
from rich.progress import Progress, BarColumn, TextColumn, TimeElapsedColumn, SpinnerColumn
from typing import Callable, List, Dict, Any, Tuple, TypeVar

from networkx.algorithms.dag import topological_sort
from networkx.classes.digraph import DiGraph
from networkx.classes.graph import NodeView

from pydantic import BaseModel

from core.constructs.models import frozendict
from core.constructs.output_manager import OutputManager, OutputTask

from core.constructs.resource import Resource_Change_Type, Resource_Difference, Resource_Reference_Difference, Resource_Reference_Change_Type, ResourceModel
from core.constructs.backend import Backend
from core.constructs.mapper import CloudMapper
from core.constructs.components import Component, Component_Change_Type, Component_Difference, ComponentModel
from core.constructs.commands import BaseCommand, BaseCommandContainer
from core.constructs.cloud_output import Cloud_Output, evaluate_dynamic_output, cloud_output_dynamic_model
from core.constructs.settings import Settings_Info, Settings, initialize_settings


from core.utils.command_finder import find_specified_command
from core.utils import module_loader, topological_helper
from core.utils.logger import log

from core.constructs.types import F


_GLOBAL_WORKSPACE: "Workspace" = None




class Workspace_Info(BaseModel):
    python_module: str
    python_class: str
    settings_info: Settings_Info
    config: Dict

    def __init__(
        __pydantic_self__, python_module: str, python_class: str, settings_info: Settings_Info, config: Dict
    ) -> None:
        """
        Represents the data needed to create a new cdev workspace:

        Parameters:
            python_module: The name of the python module to load as the workspace
            python_class: The name of the class in the python module to initialize
            settings_info: Settings_Info
            config: configuration option for the workspace

        """

        super().__init__(
            **{
                "python_module": python_module,
                "python_class": python_class,
                "settings_info": settings_info,
                "config": config,
            }
        )


class Workspace_State(str, Enum):
    UNINITIALIZED = "UNINITIALIZED"
    INITIALIZING = "INITIALIZING"
    INITIALIZED = "INITIALIZED"
    EXECUTING_FRONTEND = "EXECUTING_FRONTEND"
    EXECUTING_BACKEND = "EXECUTING_BACKEND"


def wrap_phase(phases: List[Workspace_State]) -> Callable[[F], F]:
    """
    Annotation that denotes when a function can be executed within the life cycle of a workspace. Throws excpetion if the workspace is not in the correct
    phase.
    """

    def inner_wrap(func: F) -> F:
        def wrapper_func(workspace: "Workspace", *func_posargs, **func_kwargs):

            current_state = workspace.get_state()
            if not current_state in phases:
                raise Exception(
                    f"Trying to call {func} while in workspace state {current_state} but need to be in {phases}"
                )

            else:
                rv = func(workspace, *func_posargs, **func_kwargs)
                return rv

        return wrapper_func

    return inner_wrap


class Workspace:
    """
    A singleton that encapsulates the configuration and high level information needed to construct the project. This singleton
    can be used within the different components to gain information about the higher level project that it is within. Once constructed,
    the object should remain a read only object to prevent components from changing configuration beyond their scope.
    """
    _settings: Settings

    @classmethod
    def instance(cls) -> 'Workspace':
        if not _GLOBAL_WORKSPACE:
            raise Exception("Currently No Global Workspace")

        return _GLOBAL_WORKSPACE

    @classmethod
    def set_global_instance(cls, workspace: "Workspace"):
        global _GLOBAL_WORKSPACE
        _GLOBAL_WORKSPACE = workspace

    @classmethod
    def remove_global_instance(cls, caller: "Workspace"):
        """
        Method to reset the Global Workspace object. This should be the final cleanup step for a Cdev process.
        """
        global _GLOBAL_WORKSPACE

        if not _GLOBAL_WORKSPACE:
            raise Exception("Global Workspace is not set")

        if not _GLOBAL_WORKSPACE == caller:
            raise Exception("Only the current Workspace object can remove itself")

        _GLOBAL_WORKSPACE = None

    def initialize_workspace(self, workspace_configuration: Workspace_Info):
        """
        Run the configuration needed to initialize a workspace. This should generally only be called by the Core framework itself to ensure that the
        life cycle of a workspace is correctly handled.
        """
        raise NotImplementedError

    def destroy_workspace(self):
        raise NotImplementedError

    ############################
    ##### Settings
    ############################

    
    @property
    def settings(cls) -> Settings:
        return Workspace._settings

    
    @settings.setter
    def settings(cls, value: Settings):
        Workspace._settings = value


    ############################
    ##### Initialized
    ############################
    def get_state(self) -> Workspace_State:
        """
        Get the current lifecycle state of the Workspace.

        Returns:
            state (Workspace_State)
        """
        raise NotImplementedError

    def set_state(self, value: Workspace_State):
        """
        Set the current lifecycle state of the Workspace.

        Arguments:
            value (Workspace_State)
        """
        raise NotImplementedError

    #################
    ##### Mappers
    #################
    def add_mapper(self, mapper: CloudMapper) -> None:
        """
        Add a CloudMapper to the project. The order that the Mappers are added to the Workspace defines the precedence give when
        determining which CloudMapper to use.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            mapper (CloudMapper): The mapper to add
        """
        raise NotImplementedError

    def add_mappers(self, mappers: List[CloudMapper]) -> None:
        """
        Add a List of CloudMappers to the project. The order that the Mappers are added to the Workspace defines the precedence
        give when determining which CloudMapper to use.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            mappers (List[CloudMapper]): The mapper to add
        """
        raise NotImplementedError

    def get_mappers(self) -> List[CloudMapper]:
        """
        Return the List of CloudMappers for this Workspace.

        Note that this function should only be called during the `Workspace Initialized` part of the Cdev lifecycle.

        Returns:
            mappers (List[CloudMapper]): mappers for this Workspace
        """
        raise NotImplementedError

    def get_mapper_namespace(self) -> Dict[str, CloudMapper]:
        """
        Return the Dictionary that maps Resource ID's (ruuid) to the mapper that will be used to deploy the resource into the cloud.

        Note that this function should only be called during the `Workspace Initialized` part of the Cdev lifecycle.

        Returns:
            ruuid_to_mapper (Dict[str, CloudMapper]): Resource ID to CloudMapper
        """
        raise NotImplementedError

    #################
    ##### Commands
    #################
    def add_command(self, command_location: str):
        """
        Add a Command Location to the Workspace. The order that the Command is added to the Workspace defines the precedence
        give when searching for Commands. Command Locations should adhere to the defined form to ensure that they can be
        found within the Workspace.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            command_location (Command): The command location to add
        """
        raise NotImplementedError

    def add_commands(self, command_locations: List[str]):
        """
        Add a List of Command Locations to the Workspace. The order that the Commands are added to the Workspace defines the precedence
        give when searching for a Command. Command Locations should adhere to the defined form to ensure that they can be
        found within the Workspace.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            command_locations (Command): The command location to add
        """
        raise NotImplementedError

    def get_commands(self) -> List[str]:
        """Get the Command Locations for this Workspace.

        Note that this function should only be called during the `Workspace Initialized` part of the Cdev lifecycle.

        Returns:
            command_locations (List[str])
        """
        raise NotImplementedError

    #######################
    ##### Display Output
    #######################

    def display_output(self, tag: str, output: Cloud_Output):
        """Display the output from a Resource or Reference after a process has completed

        Args:
            tag: A key value to display with the output
            output: The Cloud Output to render
        """
        raise NotImplementedError


    def render_outputs(self) -> List[Tuple[str, Any]]:
        """Render the output associated with the Workspace

        Returns:
            List[Tuple[str, Any]]: The List of outputs with their associated tag
        
        """
        raise NotImplementedError


    #######################
    ##### Components
    #######################
    def add_component(self, component: Component):
        """
        Add a Component to the Workspace. Components are used to determine the desired state of the Workspace. They should represent
        a logical separation for the Resources in a project.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            component (Component): Component to add
        """
        raise NotImplementedError

    def add_components(self, components: List[Component]):
        """
        Add a List of Components to the Workspace. Components are used to determine the desired state of the Workspace. They
        should represent a logical separation for the Resources in a project.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            component (Component): Component to add
        """
        raise NotImplementedError

    def get_components(self) -> List[Component]:
        """
        Return the Components for this Workspace.

        Note that this function should only be called during the `Workspace Initialized` or `Executing Frontend` part of the Cdev lifecycle.

        Returns:
            components (List[Component])
        """
        raise NotImplementedError

    ################
    ##### Backend
    ################
    def get_backend(self) -> Backend:
        """
        Return the Backend that is associated with this Workspace.

        Note that this function should only be called during the `Workspace Initialized` part of the Cdev lifecycle.

        Returns:
            backend (Backend): Initialized Backend for the Workspace
        """
        raise NotImplementedError

    def set_backend(self, backend: Backend):
        """
        Set the Backend for the Workspace.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            backend (Backend):  Initialized Backend for the Workspace
        """
        raise NotImplementedError

    def set_resource_state_uuid(self, resource_state_uuid: str):
        """
        Set the Resource State UUID to denote the Resource State that this Workspace will execute over.

        Note that this function should only be called during the `Workspace Initialization` part of the Cdev lifecycle.

        Arguments:
            resource_state_uuid (str): Resource State UUID from the Backend
        """
        raise NotImplementedError

    def get_resource_state_uuid(self) -> str:
        """
        Get the Resource State UUID that this Workspace is executing over.

        Note that this function should only be called during the `Workspace Initialized` part of the Cdev lifecycle.

        Returns:
            resource_state_uuid (str): Resource State UUID from the Backend
        """
        raise NotImplementedError

    @wrap_phase([Workspace_State.EXECUTING_FRONTEND])
    def generate_current_state(self) -> List[ComponentModel]:
        """
        Execute the components of this workspace to generate the current desired state of the resources.

        Returns:
            Current State (List[ComponentModel]): The current state generated by the components.
        """

        rv = []
        components: List[Component] = self.get_components()

        for component in components:
            rv.append(component.render())

        return rv

    @wrap_phase([Workspace_State.EXECUTING_FRONTEND])
    def create_state_differences(
        self, desired_state: List[ComponentModel], previous_state_component_names: List[str]
    ) -> Tuple[
        Component_Difference, Resource_Difference, Resource_Reference_Difference,
    ]:
        """
        Produce the differences between the desired state of the components and the current saved state

        Arguments:
            desired_state (List[ComponentModel]): The current state generated by the components.
            previous_state_component_names (List[str]): The names of the components to diff against.

        Returns:
            component_differences (List[Component_Difference]): The list of differences at the component level.
            reference_differences (List[Resource_Reference_Difference]): The list of differences for references within components.
            resource_differences (List[Resource_Difference]): The list of the difference for the resources within components.

        """

        backend = self.get_backend()
        resource_state_uuid = self.get_resource_state_uuid()

        return backend.create_differences(
            resource_state_uuid, desired_state, previous_state_component_names
        )

    @wrap_phase([Workspace_State.EXECUTING_FRONTEND])
    def sort_differences(self, differences: Tuple[List[Component_Difference], List[Resource_Reference_Difference], List[Resource_Difference]]) -> DiGraph:
        return topological_helper.generate_sorted_resources(differences)


    @wrap_phase([Workspace_State.EXECUTING_BACKEND])
    def deploy_differences(self, differences_dag: DiGraph) -> None:

        console = Console()
        with Progress(
                SpinnerColumn(),
                "[progress.description]{task.description}",
                BarColumn(finished_style='white'),
                TimeElapsedColumn(),
                TextColumn("{task.fields[comment]}"),
                console=console
            ) as progress:
            output_manager = OutputManager(console, progress)

            all_nodes_sorted: List[NodeView] = [x for x in topological_sort(differences_dag)]

            # There seems to be some bug in Rich that causes a deadlock when creating a bunch of Tasks in a row.
            # The problem seems to be around it refreshing after creating each task, so to avoid this problem,
            # we are disabling the console while the task are created. Then turn the console back on when they 
            # are created.
            # TODO: Recreate problem in isolated environment, and submit bug report.
            output_manager._progress.disable = True

            node_to_task= {x: output_manager.create_task(output_manager.create_output_description(x), start=False, total=10, comment='Waiting') for x in all_nodes_sorted}

            # Re-enable to console to update
            output_manager._progress.disable = False

            topological_helper.topological_iteration(differences_dag, self.wrap_output_deploy_change(node_to_task), failed_parent_handler=self.wrap_output_failed_child(node_to_task))
            


    @wrap_phase([Workspace_State.EXECUTING_BACKEND])
    def wrap_output_failed_child(self, tasks: Dict[NodeView, OutputTask]):
        def mark_failure_by_parent(change: NodeView) -> None:
            output_task = tasks.get(change)
            output_task.update(advance=10, comment="Failed because parent resource failed to deploy :cross_mark:")

        return mark_failure_by_parent


    @wrap_phase([Workspace_State.EXECUTING_BACKEND])
    def wrap_output_deploy_change(self, tasks: Dict[NodeView, OutputTask]):

        def deploy_change(change: NodeView) -> None:
            output_task = tasks.get(change)
            output_task.start_task()

            if isinstance(change, Resource_Difference):
                # Step 1 is to register a transaction with the backend

                try:
                    transaction_token, namespace_token = self.get_backend().create_resource_change_transaction(self.get_resource_state_uuid(), change.component_name, change)
                except Exception as e:
                    print(e)
                    print(f"Error Creating Transaction: {change}")
                    raise e
                

                ruuid = change.new_resource.ruuid if change.new_resource else change.previous_resource.ruuid

                # The Previous output is used by the mappers to make changes to the underlying resources.
                previous_output = (
                    self.get_backend().get_cloud_output_by_name(self.get_resource_state_uuid(), change.component_name, ruuid, change.previous_resource.name)
                    if not change.action_type == Resource_Change_Type.CREATE else
                    {}
                )

                try:
                    # Substitute the model with a model that has the cloud outputs evaluated.
                    new_evaluated_resource, evaluated_keys = self.evaluate_and_replace_cloud_output(change.component_name, change.new_resource)
                    previous_evaluated_resource = self.evaluate_and_replace_previous_cloud_output(change.component_name, change.previous_resource)

                    
                    # If there was no cloud output in the resource, then the evaluated resources will equal the original resources
                    # so using that value is safe. 
                    _evaluated_change = Resource_Difference(
                        change.action_type,
                        change.component_name,
                        previous_evaluated_resource,
                        new_evaluated_resource,
                    )

                    
                except Exception as e:
                    print(e)
                    print(f"Error evaluating cloud output from change: {change}")
                    raise e

                try:
                    log.debug("evaluated information %s", _evaluated_change)
                    output_task.update(advance=5, comment="Deploying on Cloud :cloud:")
                    mapper = self.get_mapper_namespace().get(ruuid)
                    cloud_output = mapper.deploy_resource(transaction_token, namespace_token, _evaluated_change, previous_output, output_task)
                    output_task.update(advance=3, comment="Completing transaction with Backend")

                except Exception as e:
                    self.get_backend().fail_resource_change(self.get_resource_state_uuid(), change.component_name, change, transaction_token, {"message": "deployment error"})
                    output_task.update(advance=10, comment="Failed because mapper raised error :cross_mark:")
                    print(e)
                    raise e

                try:
                    if change.action_type == Resource_Change_Type.CREATE:
                        cloud_output['cloud_region'] = Workspace.instance().settings.AWS_REGION


                    self.get_backend().complete_resource_change(
                        self.get_resource_state_uuid(),
                        change.component_name,
                        change,
                        transaction_token,
                        cloud_output,
                        evaluated_keys
                    )

                except Exception as e:
                    self.get_backend().fail_resource_change(self.get_resource_state_uuid(), change.component_name, change, transaction_token, {"message": "backend error"})
                    output_task.update(advance=10, comment="Failed because backend raised error :cross_mark:")
                    print(e)
                    raise e

            elif isinstance(change, Resource_Reference_Difference):
                output_task.update(advance=5, comment=f'{"Creating" if change.action_type == Resource_Reference_Change_Type.CREATE else "Removing"} Reference')
                self.get_backend().resolve_reference_change(self.get_resource_state_uuid(), change)


            elif isinstance(change, Component_Difference):
                output_task.update(advance=5, comment='Updating Component in Backend')
                self.get_backend().update_component(self.get_resource_state_uuid(), change)


            else:
                raise Exception(f"Trying to deploy node {change} but it is not a correct type ")


            output_task.update(completed=10, comment='Completed :white_check_mark:')


        return deploy_change

    @wrap_phase([Workspace_State.EXECUTING_BACKEND])
    def evaluate_and_replace_cloud_output(self, component_name: str, original_resource: ResourceModel) -> Tuple[ResourceModel, Dict]:
        """[stuff]

        Args:
            component_name (str): [description]
            original_resource (ResourceModel): [description]

        Returns:
            Tuple[ResourceModel, Dict]: [description]
        """
        if not original_resource:
            return original_resource, None
        
        original_resource_dict = frozendict(original_resource.dict())
        
        def _resolver(ruuid, name, key) -> Any:
            return self.get_backend().get_cloud_output_value_by_name(
                self.get_resource_state_uuid(),
                component_name,
                ruuid,
                name,
                key
            )
            
        updated_resource, resolved_value = self._recursive_replace_output(_resolver, original_resource_dict, {})
        
        evaluated_resource = original_resource.__class__(**updated_resource)

        return evaluated_resource, resolved_value


    @wrap_phase([Workspace_State.EXECUTING_BACKEND])
    def evaluate_and_replace_previous_cloud_output(self, component_name: str, previous_resource: ResourceModel) -> ResourceModel:
        """[stuff]

        Args:
            component_name (str): [description]
            original_resource (ResourceModel): [description]

        Returns:
            Tuple[ResourceModel, Dict]: [description]
        """
        
        if not previous_resource:
            return previous_resource
        
        original_resource_dict = frozendict(previous_resource.dict())
        

        def wrap_resolver() -> Callable:
            previous_resolved_values = self.get_backend().get_component(self.get_resource_state_uuid(), component_name).previous_resolved_cloud_values

            if not previous_resolved_values:
                raise Exception

            previous_resource_resolved_values = previous_resolved_values.get(f'{previous_resource.ruuid};{previous_resource.name}')

            def _resolver(ruuid, name, key) -> Any:

                if not previous_resource_resolved_values:
                    raise Exception

                key  = f"{ruuid};{name};{key}"

                if not key in previous_resource_resolved_values:
                    raise Exception

                return previous_resource_resolved_values.get(key)

            return _resolver
            
            
        updated_resource, _ = self._recursive_replace_output(wrap_resolver(), original_resource_dict, {})
        
        evaluated_resource = previous_resource.__class__(**updated_resource)

        return evaluated_resource


    def _recursive_replace_output(self, resolver: Callable, original: Any, resolved_values: Dict = {}) -> Tuple[Any, Dict]:
        # This function works over ImmutableModel that were converted using the `.dict` method. Therefore,
        # the collections will be `frozendict` and `frozenset` instead of `dict` and `list`
        
        if isinstance(original, frozendict):
            if "id" in original and original.get("id") == 'cdev_cloud_output':
                
                _value = resolver(
                    original.get('ruuid'),
                    original.get('name'),
                    original.get('key')
                )
                
                values_key = f"{original.get('ruuid')};{original.get('name')};{original.get('key')}"
                resolved_values[values_key] = _value

                if original.get('output_operations'):
                    return (evaluate_dynamic_output(_value, cloud_output_dynamic_model(**original)), resolved_values)

                return (_value, resolved_values)

            rv = {}
            for k,v in original.items():
                new_v, tmp_resolved_values = self._recursive_replace_output(resolver, v, resolved_values)
                
                rv[k] = new_v

                resolved_values.update(tmp_resolved_values)


            return (frozendict(rv), resolved_values)

        elif isinstance(original, frozenset): 
            rv = []
            for x in original:
                new_v, tmp_resolved_values = self._recursive_replace_output(resolver, x)
                
                rv.append(new_v)

                resolved_values.update(tmp_resolved_values)

            return frozenset(rv), resolved_values

        
        return original, resolved_values


    @wrap_phase([Workspace_State.INITIALIZED])
    def execute_command(self, command: str, args: List):
        """
        Find the desired command based on the search path and execute it with the given arguments.

        Args:
            command (str): The full command to search for. can be '.' seperated to denote search path.
            args (List): The command lines arguments to pass to the command.

        Raises:
            KeyError: Raises an exception.
        """

        # Command in list form
        command_list = command.split(".")

        # Create list of all directories to start searching in
        all_search_locations_list = self.get_commands()


        obj, program_name, command_name, is_command = find_specified_command(
            command_list, all_search_locations_list
        )

        if is_command:
            if not isinstance(obj, BaseCommand):
                raise Exception

            try:
                args = [program_name, command_name, *args]
                obj.run_from_command_line(args)
            except Exception as e:
                raise e

        else:
            if not isinstance(obj, BaseCommandContainer):
                raise Exception

            try:
                obj.display_help_message()
            except Exception as e:
                raise e


class WorkspaceManager:
    def create_new_workspace(self, workspace_info: Workspace_Info, *posargs, **kwargs):
        """
        Create a new workspace based on the information provided.

        Args:
            workspace_info (Workspace_Info): information about the backend configuration

        Raises:
            WorkSpaceAlreadyCreated
        """
        raise NotImplementedError

    def check_if_workspace_exists(self, *posargs, **kwargs) -> bool:
        raise NotImplementedError

    def load_workspace_configuration(self, *posargs, **kwargs) -> Workspace_Info:
        raise NotImplementedError

    def load_workspace(self, *posargs, **kwargs) -> Workspace:
        raise NotImplementedError


def load_and_initialize_workspace(config: Workspace_Info):
    """
    Load and initialize the workspace from the given configuration
    """
    ws = load_workspace(config)
    ws.set_state(Workspace_State.INITIALIZING)
    initialize_workspace(ws, config.settings_info, config.config)
    ws.set_state(Workspace_State.INITIALIZED)


def load_workspace(config: Workspace_Info) -> Workspace:
    """
    Load and initialize the workspace from the given configuration
    """
    try:
        workspace_module = module_loader.import_module(config.python_module)
    except Exception as e:
        print("Error loading workspace module")
        print(f"Error > {e}")

        raise e

    workspace_class = None
    for item in dir(workspace_module):
        potential_obj = getattr(workspace_module, item)
        if (
            inspect.isclass(potential_obj)
            and issubclass(potential_obj, Workspace)
            and item == config.python_class
        ):
            workspace_class = potential_obj
            break

    if not workspace_class:
        print(f"Could not find {config.python_class} in {config.python_module}")
        raise Exception

    try:
        # initialize the backend obj with the provided configuration values
        return workspace_class()

    except Exception as e:
        print(
            f"Could not load {workspace_class} Class from config {config.config}; {e}"
        )
        raise e


def initialize_workspace(workspace: Workspace, settings: Settings_Info , config: Dict):
    try:
        Workspace.settings = initialize_settings(settings)
        
        # initialize the backend obj with the provided configuration values
        workspace.initialize_workspace(config)
        

    except Exception as e:
        print(
            f"Could not initialize {workspace} Class from config {config}; {e}"
        )
        raise e







"""Utilities for initializing the workspace

"""


import os

from ..constructs.workspace import Workspace_Info

from ..constructs.workspace import load_and_initialize_workspace
from ..default.workspace import local_workspace_manager


def initialize_workspace_cli(args):

    workspace_manager = local_workspace_manager(os.getcwd())

    workspace_config = workspace_manager.load_workspace_configuration()

    try:
        initialize_workspace(workspace_config)
    except Exception as e:
        raise e


def initialize_workspace(workspace_config: Workspace_Info):

    try:
        load_and_initialize_workspace(workspace_config)
    except Exception as e:
        raise e



"""Set of constructs for creating Cloud Object Stores like S3

"""

from enum import Enum
from typing import Any, List

from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs, PermissionsAvailableMixin
from core.constructs.cloud_output import Cloud_Output_Str, OutputType
from core.constructs.types import cdev_str_model
from core.utils import hasher

from core.default.resources.simple.iam import Permission

from core.default.resources.simple.events import Event, event_model

RUUID = "cdev::simple::bucket"

class Bucket_Event_Type(str, Enum):
    Object_Created = "s3:ObjectCreated:*"
    """Receive Events when Objects are created in the `Bucket`"""
    Object_Removed = "s3:ObjectRemoved:*"
    """Receive Events when Objects are removed from the `Bucket`"""
    Object_Restore = "s3:ObjectRestore:*"
    """Receive Events when Objects are restored to the `Bucket`"""
    Object_Replicaton = "s3:ObjectReplication:*"
    """Receive Events when Objects are replicated to the `Bucket`"""


class bucket_event_model(event_model):
    """Model representing an Event from a Bucket

    Args:
        bucket_arn (cdev_str_model)
        bucket_name (cdev_str_model)
        bucket_event_type (Bucket_Event_Type)   
    """
    bucket_arn: cdev_str_model
    bucket_name: cdev_str_model
    bucket_event_type: Bucket_Event_Type


class BucketEvent(Event):
    """Construct representing an Event from a Bucket
    """

    def __init__(self, bucket_name: str, bucket_event_type: Bucket_Event_Type) -> None:
        """
        Args:
            bucket_name (str): Cdev name of the bucket
            bucket_event_type (Bucket_Event_Type): Type fo events to create
        """
        self.bucket_cdev_name = bucket_name
        self.bucket_event_type = bucket_event_type
        self.bucket_arn =  Cloud_Output_Str(
                name=bucket_name, 
                ruuid=RUUID, 
                key='cloud_id',
                type=OutputType.RESOURCE 
            )
        self.bucket_name =  Cloud_Output_Str(
                name=bucket_name, 
                ruuid=RUUID, 
                key='bucket_name',
                type=OutputType.RESOURCE 
            )

    def render(self) -> bucket_event_model:
        return bucket_event_model(
            originating_resource_name=self.bucket_cdev_name,
            originating_resource_type=RUUID,
            hash=self.hash(),
            bucket_event_type=self.bucket_event_type,
            bucket_arn=self.bucket_arn.render(),
            bucket_name=self.bucket_name.render()
        )

    def hash(self) -> str:
        return hasher.hash_list([
            self.bucket_cdev_name,
            self.bucket_event_type
        ])


class BucketPermissions:
    RUUID = "cdev::simple::bucket"

    def __init__(self, resource_name: str) -> None:
        self.READ_BUCKET = Permission(
            actions=[
                "s3:GetObject",
                "s3:GetObjectVersion", 
                "s3:ListBucket"
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to read objects from the `Bucket`"""

        self.WRITE_BUCKET = Permission(
            actions=[
                "s3:PutObject",
                "s3:PutObjectAcl", 
                "s3:ListBucket"
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to write objects to the `Bucket`"""

        self.READ_AND_WRITE_BUCKET = Permission(
            actions=[
                "s3:*Object",
                "s3:ListBucket"
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to read and write objects to and from the `Bucket`"""

        self.READ_EVENTS = Permission(
            actions=[
                "s3:*Object",
                "s3:ListBucket"
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to receive events from the `Bucket`"""


class bucket_model(ResourceModel):
    """Model representing a Bucket"""
    pass


class BucketOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after a Bucket has been deployed."""
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)


    @property
    def bucket_name(self) -> Cloud_Output_Str:
        """The name of the generated bucket"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='bucket_name',
            type=self.OUTPUT_TYPE
        )

    @bucket_name.setter
    def bucket_name(self, value: Any):
        raise Exception


class Bucket(PermissionsAvailableMixin, Resource):
    """A Simple Bucket is a basic object store for applications to build on."""
    
    @update_hash
    def __init__(self, cdev_name: str, nonce: str="") -> None:
        """
        Args:
            cdev_name (str): Name of the resource
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.
        """
        super().__init__(cdev_name, RUUID, nonce)

        self.available_permissions: BucketPermissions = BucketPermissions(cdev_name)
        self.output = BucketOutput(cdev_name)


    def create_event_trigger(
        self, event_type: Bucket_Event_Type
    ) -> BucketEvent:
        """Create Construct of event that other resources can respond to"""

        event = BucketEvent(
            bucket_name=self.name,
            bucket_event_type=event_type
        )

        return event

    def compute_hash(self):
        self._hash = hasher.hash_list([self.nonce])
        
    def render(self) -> bucket_model:
        return bucket_model(
            ruuid=self.ruuid,
            name=self.name,
            hash=self.hash,
        )




import os
import hashlib
from typing import List
from pydantic.types import DirectoryPath, FilePath

from sortedcontainers import SortedDict

from .parser_objects import *
from .import parser_utils as p_utils
from .parser_exceptions import *



## This file has the highest level functions for use by other modules. Only this file will expose functionality to other modules.


def parse_functions_from_file(file_loc, include_functions=[], function_manual_includes={}, global_manual_includes=[], remove_top_annotation=False) -> file_information:
    try:
        file_information = p_utils.get_file_information(
            file_loc, include_functions, function_manual_includes, global_manual_includes, remove_top_annotation)
    except InvalidDataError as e:
        file_information = None
        print(f"{type(e)} -> {e.message}")
    except CouldNotParseFileError as e:
        file_information = None
        print(f"{type(e.error)} -> {e.error.message}")

    return file_information



def parse_file_for_dependencies(file_loc: DirectoryPath):
    return p_utils.get_file_imported_symbols(file_loc)



"""Structure that encapsulates a namespace of deploy resources

Properties of a component:
[]
"""

from enum import Enum
from typing import Dict, Union, List, Optional, Set


from pydantic import BaseModel

from .resource import ResourceReferenceModel, ResourceModel


class ComponentModel(BaseModel):
    """

    This is the most basic information needed to describe a rendered component.

    --- Attributes ---

    - resources ->  a list of rendered resources that make up this resource

    - hash  ->  a string that is the hash of this component. This hash must be computed such that
                it changes only if a change in the state is desired.

    - name  ->  a string that is the human readable name for this component
    """

    name: str
    """
    A human readable logical name for the component. 

    This value is important for allow human level refactoring of components. To update a component name once created, you must edit only the 
    name and not change the hash. If you change both the hash and name in the same deployment, it will register this as a delete and create 
    instead of update. 
    """


    hash: str
    """
    A hash that is used to identify if changes in the resources have occurred. It should have the property:
    - This value changes only if a there is a change in the resource. 
    """

    resources: Optional[List[ResourceModel]]
    """
    List of Rendered Resources that make up the current state of this component
    """

    references: Optional[List[ResourceReferenceModel]]
    """
    A list of the referenced resources used by this component
    """

    cloud_output: Optional[Dict[str, Dict]]
    """
    Output values from the cloud provider of deployed resources
    """

    previous_resolved_cloud_values: Optional[Dict[str, Dict]]
    """
    Previously resolved cloud output
    """

    external_references: Optional[Dict[str, Dict]]
    """
    Dictionary of the resources to information about references to them from other components
    """

    def __init__(
        __pydantic_self__,
        name: str,
        hash: str = "0",
        resources: List[ResourceModel] = [],
        references: List[ResourceReferenceModel] = [],
        cloud_output: Dict[str, Dict] = {},
        previous_resolved_cloud_values: Dict[str, Dict]={},
        external_references: Dict[str, Dict] = {},
    ) -> None:
        super().__init__(
            **{
                "name": name,
                "hash": hash,
                "resources": list(resources),
                "references": list(references),
                "cloud_output": dict(cloud_output),
                "previous_resolved_cloud_values": dict(previous_resolved_cloud_values),
                "external_references": dict(external_references),

            }
        )


class Component_Change_Type(str, Enum):
    CREATE = "CREATE"
    UPDATE_NAME = "UPDATE_NAME"
    DELETE = "DELETE"
    UPDATE_IDENTITY = "UPDATE_IDENTITY"


class Component_Difference(BaseModel):
    action_type: Component_Change_Type
    previous_name: Optional[str]
    new_name: Optional[str]

    def __init__(
        __pydantic_self__,
        action_type: Component_Change_Type,
        previous_name: str = None,
        new_name: str = None,
    ) -> None:
        super().__init__(
            **{
                "action_type": action_type,
                "previous_name": previous_name,
                "new_name": new_name,
            }
        )

    class Config:
        use_enum_values = True
        # Beta Feature but should be fine since this is simple data 
        frozen = True



class Component:
    """
    A component is a logical collection of resources. This simple definition is intended to allow flexibility for different
    styles of setup. It is up to the end user to decide on how they group the resources.

    A component must override the `render` method, which returns the desired resources with configuration as a component model.
    The `render` method does not take any input parameters, therefore all configuration for the component should be done via the `__init__`
    method or other defined methods.
    """

    def __init__(self, name: str):
        self.name = name

    def render(self) -> ComponentModel:
        """Abstract Class that must be implemented by the descendant that returns a component model"""
        raise NotImplementedError

    def get_name(self) -> str:
        return self.name



from typing import Any, Dict, FrozenSet, List, Tuple
from uuid import uuid4
from core.constructs.models import frozendict

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.constructs.output_manager import OutputTask
from core.default.resources.simple import api as simple_api

from .. import aws_client 


class NoAuthorizerIdFoundError(Exception):
    pass


def _create_simple_api(
        transaction_token: str, 
        namespace_token: str, 
        resource: simple_api.simple_api_model, 
        output_task: OutputTask
    ) -> Dict:
    """
    Create an API on AWS.

    Args:
        transaction_token (str): Transaction token to use to identify this action over a resource.
        namespace_token (str): Token used to make sure that resources are created properly in AWS.
        resource (simple_api.simple_api_model): The information about what should be deployed.
        output_task (OutputTask): Output Task to send any progress information.

    Raises:
        e: [description]

    Returns:
        Dict: Information from the cloud about the resource.
    """
    
    base_args = {
        "Name": f"cdev-api-{namespace_token}-{str(uuid4())}",
        "ProtocolType": "HTTP",
    }

    cors_args = {
        "CorsConfiguration": {
            "AllowOrigins": ["*"],
            "AllowMethods": ["*"],
            "AllowHeaders": [
                "Content-Type",
                "X-Amz-Date",
                "Authorization",
                "X-Api-Key",
                "X-Amz-Security-Token",
                "X-Amz-User-Agent",
            ],
        }
    }

    if resource.allow_cors:
        base_args.update(cors_args)

    output_task.update(advance=1, comment='Creating Api')

    try:
        rv = aws_client.run_client_function("apigatewayv2", "create_api", base_args)
    except Exception as e:
        output_task.print_error(e)
        raise e 

    api_id = rv.get("ApiId")

    info = {
        "cloud_id": api_id,
        "endpoints": {},
    }

    
    if resource.authorizers:
        _authorizer_outputs: Dict[str, Dict] = {}
        
        for authorizer in resource.authorizers:
            output_task.update(advance=1, comment=f'Creating Authorizer {authorizer.name}')
            authorizer_id = _create_authorizer(api_id, authorizer)
            _authorizer_outputs[authorizer_id] = authorizer.dict()

        info['authorizers'] = _authorizer_outputs
        
    

    _stage_args = {
        "ApiId": info.get("cloud_id"),
        "AutoDeploy": True,
        "StageName": "live",
    }
    
    output_task.update(advance=1, comment='Creating Stage')
    try:
        rv2 = aws_client.run_client_function('apigatewayv2', 'create_stage', _stage_args)
    except Exception as e:
        output_task.print_error(e)
        raise e 

    info["endpoint"] = f"{rv.get('ApiEndpoint')}/{rv2.get('StageName')}"


    api_id = info.get("cloud_id")
    if resource.routes:
        _created_endpoints = {}
        for route in resource.routes:
            output_task.update(advance=1, comment=f'Creating Route {route.path} [{route.verb}]')
            
            try:
                authorizer_id = _find_authorization_id(route, info.get('authorizers') )

                route_cloud_id = _create_route(api_id, route, authorizer_id)
            except Exception as e:
                output_task.print_error(e)
                raise e 


            # Add route to the return info
            dict_key = f'{route.path} {route.verb}'

            _created_endpoints[dict_key] = route_cloud_id

        
        info["endpoints"] = _created_endpoints


    return info


def _update_simple_api(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: simple_api.simple_api_model,
    new_resource: simple_api.simple_api_model,
    previous_output: frozendict,
    output_task: OutputTask
    ) -> Dict:
    """
    Create an API on AWS.

    Args:
        transaction_token (str): Transaction token to use to identify this action over a resource.
        namespace_token (str): Token used to make sure that resources are created properly in AWS.
        previous_resource (simple_api.simple_api_model): The information about the previous API.
        new_resource (simple_api.simple_api_model): The information about the new API.
        previous_output (Dict): Cloud output of the previous resource.
        output_task (OutputTask): Output Task to send any progress information.

    Raises:
        e: [description]

    Returns:
        Dict: Information from the cloud about the resource.
    """

    mutable_previous_output = dict(previous_output)
    previous_cloud_id = mutable_previous_output.get('cloud_id')

    output_task.print(previous_resource)

    
    # Change the CORS Settings of the API
    if not previous_resource.allow_cors == new_resource.allow_cors:
        new_cors_policy =  {
            "AllowOrigins": ["*"],
            "AllowMethods": ["*"],
            "AllowHeaders": [
                "Content-Type",
                "X-Amz-Date",
                "Authorization",
                "X-Api-Key",
                "X-Amz-Security-Token",
                "X-Amz-User-Agent",
            ],
        } if new_resource.allow_cors else {}

        output_task.update(advance=1, comment=f'Updating CORS Policy')
        try:
            aws_client.run_client_function(
                'apigatewayv2', 
                'update_api', 
                {
                    'api_id': previous_cloud_id,
                    'CorsConfiguration ': new_cors_policy
                }
            )
        except Exception as e:
            output_task.print_error(e)
            raise e 

    # If an authorizer is to be deleted, it must already be removed from any route, but to create a route with an authorizer, you must 
    # have already created the authorizer. This means we can NOT do all the operations on the routes before authorizers but also can NOT do all
    # the authorizers before the routes. 

    # We must do them in an order that does not cause errors, so we are only going to perform update operations that must happen before the 
    # authorizers are updated then defer the rest of the updates till after by storing their info in the `_update_route_info` and `_create_route_info` 
    # list.
    _create_route_info: List[simple_api.route_model] = []
    _update_route_info: List[Tuple[str, simple_api.route_model]] = []
    previous_route_info: Dict[str,str] = dict(mutable_previous_output.get('endpoints')) if mutable_previous_output.get('endpoints') else {}
    new_output_info = {}

    if not previous_resource.routes == new_resource.routes:
        
        update_routes = set()

        # Delete any route that is not in the new routes but in previous routes
        routes_to_be_deleted: FrozenSet[simple_api.route_model] = previous_resource.routes.difference(new_resource.routes)
        # Create any route that is in the new routes but not in previous routes
        routes_to_be_created: FrozenSet[simple_api.route_model] = new_resource.routes.difference(previous_resource.routes)

        previous_route_ids = set([f"{x.path} {x.verb}" for x in previous_resource.routes])

        for route in routes_to_be_created:
            route_id = (
                f'{route.path} {route.verb}'
            )

            if route_id in previous_route_ids:
                # This is updating the routes authorization not making a new route
                # There are a few states that require making an update to the current route before the authorizers are updated.
                # Updating to a new Authorizer
                #      - Previous Authorizer was none -> No update needed
                #      - Previous Authorizer existed -> remove the authorizer from the route 

                update_routes.add(route_id)

                route_cloud_id = previous_route_info.get(route_id)

                if not mutable_previous_output.get('authorizers'):
                    # No previous authorization info means all route updates should wait til authorizations 
                    # have completed
                    _update_route_info.append((route_cloud_id, route))
                    continue

                # guranteed to have an element because the invariant of the if statement
                previous_route = [x for x in previous_resource.routes if f"{x.path} {x.verb}" == route_id][0]

                # Find the previous id... we have to pass in the previous api since that was used in the previous computation regarding
                # the default authorizer
                try:
                    previous_authorizer_id = _find_authorization_id(previous_route, mutable_previous_output.get('authorizers') )
                except NoAuthorizerIdFoundError as e:
                    # If a previous authorizer was to be found and it wasn't then there is something wrong since we should have all the info on previous
                    # authorizers
                    raise e
                
                if previous_authorizer_id:
                    # remove the previous authorizer, but defer the updating to the new one until after authorizers have finished
                    print(f'soft update {route}')
                    _update_route(previous_cloud_id, route_cloud_id, route, None)
                    _update_route_info.append((route_cloud_id, route))

                else:
                    # 1A 
                    # No previous authorizer so wait til after to complete the authorization
                    _update_route_info.append((route_cloud_id, route))

            else:
                # All creates should just happen after the authorizers have been made
                _create_route_info.append(route)
                    
    
        for route in routes_to_be_deleted:
            # All deletes should go ahead and occur now
            route_id = (
                f'{route.path} {route.verb}'
            )

            if route_id in update_routes:
                continue

            output_task.update(advance=1, comment=f'Deleting Route {route.path} [{route.verb}]')
            try:
                _delete_route(
                    previous_cloud_id, previous_route_info.get(route_id)
                )
            except Exception as e:
                output_task.print_error(e)
                raise e 

            previous_route_info.pop(route_id)

    if not previous_resource.authorizers == new_resource.authorizers:
        # Three options for types of changes to authorizers
        # 1. Complete New 
        # 2. Complete Remove
        # 3. Update: This will show up as a delete and create when doing the set difference, so when going through the creates we 
        #    should look through the past authorizers to find one with the same name. 
        new_authorizer_info = {}
        previous_authorizers_info = dict(mutable_previous_output.get('authorizers')) if mutable_previous_output.get('authorizers') else {}
        updated = set()



        authorizers_to_delete: FrozenSet[simple_api.authorizer_model] = previous_resource.authorizers.difference(new_resource.authorizers)
        authorizers_to_create: FrozenSet[simple_api.authorizer_model] = new_resource.authorizers.difference(previous_resource.authorizers)

        for authorizer in authorizers_to_create:
            if any(x.name == authorizer.name for x in authorizers_to_delete):
                #update not hard create
                output_task.update(advance=1, comment=f'Updating Authorizer {authorizer.name}')
                
                authorizer_id = [id for id, v in previous_authorizers_info.items() if v.get("name") == authorizer.name][0]
                _update_authorizer(previous_output.get('cloud_id'), authorizer_id, authorizer)

                # Add this to updated authorizers so that it does not delete the authorizer in next steps
                updated.add(authorizer.name)
                # In the previous output, update the authorization info
                new_authorizer_info[authorizer_id] = authorizer.dict()

            else:
                output_task.update(advance=1, comment=f'Creating Authorizer {authorizer.name}')
                authorizer_id = _create_authorizer(previous_output.get('cloud_id'), authorizer)
                new_authorizer_info[authorizer_id] = authorizer.dict()

        

        for authorizer in authorizers_to_delete:
            if authorizer.name in updated:
                continue

            authorizer_id = [id for id, v in previous_authorizers_info.items() if v.get("name") == authorizer.name][0]
            output_task.update(advance=1, comment=f'Deleteing Authorizer {authorizer.name}')
            _delete_authorizer(previous_output.get('cloud_id'), authorizer_id)

            previous_authorizers_info.pop(authorizer_id)


        previous_authorizers_info.update(new_authorizer_info)
        mutable_previous_output['authorizers'] = previous_authorizers_info


    
    for route in _create_route_info:
        # Now that all updates to the authorizers have completed, we can do the create routes
        route_id = (
            f'{route.path} {route.verb}'
        )

        # Find the authorizer id 
        authorizer_id = _find_authorization_id(route, mutable_previous_output.get('authorizers') )

        output_task.update(advance=1, comment=f'Creating Route {route.path} [{route.verb}]')

        route_cloud_id = _create_route(previous_cloud_id, route, authorizer_id)
        
        new_output_info[route_id] = route_cloud_id


    for id, route in _update_route_info:
        # Now that all updates to the authorizers have completed, we can do the update routes that depends on the created
        # authorization
        route_id = (
            f'{route.path} {route.verb}'
        )

        # Find the authorizer id 
        authorizer_id = _find_authorization_id(route, mutable_previous_output.get('authorizers'))

        output_task.update(advance=1, comment=f'Updating Route {route.path} [{route.verb}]')

        _update_route(previous_cloud_id, id, route, authorizer_id)

    previous_route_info.update(new_output_info)
    mutable_previous_output['endpoints'] = previous_route_info

    return mutable_previous_output


def _remove_simple_api(
    transaction_token: str,  
    previous_output: Dict, 
    output_task: OutputTask
    ):
    """
    Delete an API from AWS.

    Args:
        transaction_token (str): Transaction token to use to identify this action over a resource.
        previous_output (Dict): Cloud output of the resource. Used to determine what API to delete on AWS.
        output_task (OutputTask): Output Task to send any progress information.

    Raises:
        e: [description]
    """

    api_id = previous_output.get("cloud_id")

    output_task.update(advance=1, comment=f'Deleting API')

    try:
        aws_client.run_client_function("apigatewayv2", "delete_api", {"ApiId": api_id})
    except Exception as e:
        output_task.print_error(e)
        raise e 


def _create_authorizer(api_id: str, authorizer: simple_api.authorizer_model) -> str:
    """Helper function for creating new authorizers

    Args:
        api_id (str): Api ID of the api in AWS.
        authorizer (simple_api.authorizer_model):Information about the authorizer to create

    Returns:
        str: Authorizer Id of the created Authorizer
    """
    args = {
        "ApiId": api_id,
        "Name": authorizer.name,
        "AuthorizerType": 'JWT',
        "IdentitySource":[
            '$request.header.Authorization',
        ],
        "JwtConfiguration": {
            'Audience': [
                authorizer.audience,
            ],
            "Issuer": authorizer.issuer_url
        }
    }

    rv = aws_client.run_client_function('apigatewayv2', 'create_authorizer', args)

    return rv.get("AuthorizerId")


def _update_authorizer(api_id: str, authorizer_id: str, authorizer: simple_api.authorizer_model):
    args = {
        "ApiId": api_id,
        "AuthorizerId": authorizer_id,
        "JwtConfiguration": {
            'Audience': [
                authorizer.audience,
            ],
            "Issuer": authorizer.issuer_url
        }
    }

    aws_client.run_client_function('apigatewayv2', 'update_authorizer', args)


def _delete_authorizer(api_id: str, authorizer_id: str):
    args = {
        "ApiId": api_id,
        "AuthorizerId": authorizer_id
    }
    aws_client.run_client_function('apigatewayv2', 'delete_authorizer', args)



def _find_authorization_id(route: simple_api.route_model, output_ids: Dict[str, Dict]) -> str:
    """Function for finding a route's authorizer cloud id

    This function takes into account that a route will by default use the api's default authorizer unless the route has the `override_authorizer_name`
    set. Will return None if the route will not have an authorizer or there is not enough information to find the authorizer. 

    Args:
        api_resource (simple_api.simple_api_model): _description_
        route (simple_api.route_model): _description_
        output_ids (Dict[str, Dict]): _description_

    Raises:
        Exception: _description_
        Exception: _description_

    Returns:
        str: _description_
    """

    if not route.authorizer_name:
        return None

    if not output_ids:
        return None

    
    found_id = [id for id, x in output_ids.items() if x.get("name") == route.authorizer_name]


    if len(found_id) == 0 and route.authorizer_name:
        # We have an authorizer but can not find the info for it
        raise NoAuthorizerIdFoundError

    if len(found_id) > 1:
        raise Exception

    return found_id[0]

    


def _create_route(api_id: str, route: simple_api.route_model, authorizer_id: str=None) -> str:
    """
    Helper Function for creating routes on an API. Note that any error raised by the aws client will not be caught by this function and should
    be handled by the caller of this function.

    Args:
        api_id (str): Api ID of the api in AWS.
        route (simple_api.route_event_model): Information about the route to create.

    Returns:
        str: Route ID of the create route.
    """

    full_args = {}
    _authorizer_args = {}

    _route_args = {
        "ApiId": api_id,
        "RouteKey": f"{route.verb} {route.path}",
    }

    if authorizer_id:
        _authorizer_args['AuthorizerId'] = authorizer_id
        _authorizer_args['AuthorizationType'] = 'JWT'
        
        if route.additional_scopes:
            _authorizer_args['AuthorizationScopes'] = list(route.additional_scopes)
    

    full_args.update(_route_args)
    full_args.update(_authorizer_args)
    rv = aws_client.run_client_function('apigatewayv2', 'create_route', full_args)

    return rv.get("RouteId")


def _delete_route(api_id: str, route_id: str):
    """
    Helper Function for deleting routes on an API. Note that any error raised by the aws client will not be caught by this function and should
    be handled by the caller of this function.

    Args:
        api_id (str): Api ID of the api in AWS.
        route_id (lambda_event): Route ID of the route in AWS.
    """

    aws_client.run_client_function(
        "apigatewayv2", "delete_route", {"ApiId": api_id, "RouteId": route_id}
    )


def _update_route(api_id: str, route_id: str, route: simple_api.route_model, authorizer_id: str=None):

    args = {
        "ApiId": api_id,
        "RouteId": route_id,
    }

    if authorizer_id:
        args['AuthorizerId'] = authorizer_id
        args['AuthorizationType'] = 'JWT'

        if route.additional_scopes:
            args['AuthorizationScopes'] = list(route.additional_scopes)

    else:
        args['AuthorizerId'] = ""
        args['AuthorizationType'] = "NONE"
        args['AuthorizationScopes'] = []

    aws_client.run_client_function('apigatewayv2', 'update_route', args)
    


def handle_simple_api_deployment(
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
        ) -> Dict:
    
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_api(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )

    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
        return _update_simple_api(
            transaction_token,
            namespace_token,
            simple_api.simple_api_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.DELETE:

        _remove_simple_api(
            transaction_token,
            previous_output,
            output_task
        )

        return {}




"""Structure that encapsulates the desire to capture and use output from deploying a resource on the cloud.

Often, it is required to use the returned values provided by the cloud as input into another resource. For 
example, wanting to pass a created db name into the environment variables of a serverless function that will
reference it. Therefor, it is important that these values are core primitives within the framework.

The core framework has tried to make working with these types of values feel as natural as possible, but it
is important to have an understanding of how they are designed to avoid confusion. A `Cloud Output` represents
a future value that is not currently know at the execution time of the current code. They are designed to look
and feel like the `types` they will evaluate to by providing methods that mirror common methods of the evaluated
type.

For example, a `Cloud Output String` contains methods such as `replace` that mirror `replace` on a regular string.
These methods can be chained together to create more complex functionality. For a more in depth discussion on the 
capabilities and limits of this system read out documentation at <link>.

"""

from enum import Enum
from typing import Any, List, Tuple, NewType, overload, Optional, Union, Iterable, Mapping, TypeVar, Generic
from typing_extensions import Literal, SupportsIndex
from collections.abc import Sequence

from core.constructs.models import ImmutableModel, frozendict
from core.utils import hasher

CLOUD_OUTPUT_ID = 'cdev_cloud_output'

# Wrapper type to help keep annotations compact
output_operation = NewType('output_operation', Tuple[str, Tuple, frozendict])

class OutputType(str, Enum):
    """Type of Cloud Output 
    
    Since Cloud Output can be derived from resources and references, we need to denote where the Output is coming from.
    """
    RESOURCE = 'resource'
    REFERENCE = 'reference'

########################
##### Immutable Models
########################
class cloud_output_model(ImmutableModel):
    """
    Often we want resources that depend on the value of output of other resources that is only known after a cloud resource is created. This serves
    as a placeholder for that desired value until it is available.
    """

    name: str
    """
    Name of the resource
    """

    ruuid: str
    """
    Ruuid of the resource
    """

    key: str
    """
    The key to lookup the output value by (ex: arn)
    """

    type: OutputType
    """
    Type of the underlying item we want to get the output of
    """

    id: Literal['cdev_cloud_output']
    """
    Literal that allows a dict to be identified as a Cloud Output Model
    """


class cloud_output_dynamic_model(cloud_output_model):
    """
    A cloud output model that has accompanying operations to be applied to the derived value
    """

    output_operations: Tuple[output_operation,...]
    """
    Tuple of all the operations to be applied to the derived value
    """



def evaluate_dynamic_output(original_value: Any, cloud_output_dynamic: cloud_output_dynamic_model) -> Any:
    """Evaluate a set of operations on a starting value.

    Args:
        original_value (Any): The original value to operate on
        cloud_output_dynamic (cloud_output_dynamic_model): The model containing the operations to execute

    Raises:
        Exception: [description]

    Returns:
        Any: The transformed value
    """
    operations = cloud_output_dynamic.output_operations

    intermediate_value = original_value
    for x in operations:
        func_name = x[0]
        xargs = x[1]
        kwargs = x[2]

        if func_name == '**not':
            # There is not hidden method that implements not, so we need to hard code this.
            new_rv = not intermediate_value

        elif func_name == "**and":
            new_rv = intermediate_value and xargs[0]

        elif func_name == "**or":
            new_rv = intermediate_value or xargs[0]

        elif func_name == "**xor":
            new_rv = intermediate_value ^ xargs[0]

        elif func_name == "join":
            new_rv = getattr(intermediate_value, func_name)(xargs)


        else:
            object_methods = set([method_name for method_name in dir(str) if callable(getattr(str, method_name))])
    
    
            if not func_name in object_methods:
                print(object_methods)
                raise Exception(f"'{func_name}' not in available methods for {intermediate_value} ({type(intermediate_value)})")
    
            if xargs and kwargs:
                new_rv = getattr(intermediate_value, func_name)(**kwargs)
            elif (not xargs) and kwargs:
                new_rv = getattr(intermediate_value, func_name)(**kwargs)
            elif xargs and (not kwargs):
                new_rv = getattr(intermediate_value, func_name)(*xargs)
            elif (not xargs) and (not kwargs):
                new_rv = getattr(intermediate_value, func_name)() 
    
    
        intermediate_value = new_rv


    return intermediate_value


########################
##### Helper Classes
########################
class Cloud_Output():
    """
    Mutable Class that can used during the creation phases to represent a desired cloud output model.
    """


    def __init__(self, name: str, ruuid: str, key: str, type: OutputType) -> None:
        self._name = name
        self._ruuid = ruuid
        self._key = key
        self._type = type


    def render(self) -> cloud_output_dynamic_model:
        return cloud_output_model(
            name=self._name,
            ruuid=self._ruuid,
            key=self._key,
            type=self._type,
            id='cdev_cloud_output',
        )

    def hash(self) -> str:
        return hasher.hash_list([
            self._name,
            self._ruuid,
            self._key,
            self._type
        ])

class Cloud_Output_Dynamic(Cloud_Output):
    """
    Mutable Class that can used during the creation phases to represent a desired cloud output model. Allows the user to define
    a list of operations that should be applied to the retrieved value. 
    """
    def __init__(self, name: str, ruuid: str, key: str, type: OutputType) -> None:
        super().__init__(name, ruuid, key, type)
        self._operations: List[output_operation] = []

    def render(self) -> cloud_output_dynamic_model:
        operations = tuple([(x[0], tuple(x[1]), frozendict(x[2])) for x in self._operations])

        return cloud_output_dynamic_model(
            name=self._name,
            ruuid=self._ruuid,
            key=self._key,
            type=self._type,
            id='cdev_cloud_output',
            output_operations=operations
        )


    def hash(self) -> str:
        return hasher.hash_list([
            super().hash(),
            self.render().output_operations if self._operations else ""
        ])

   


########################
##### Types of Output
########################
    
class Cloud_Output_Bool(Cloud_Output_Dynamic):
    """
    Cloud Output that will resolve to a Boolean value after being retrieve or after all the operations have been executed
    """

    def __init__(self, name: str, ruuid: str, key: str, type: OutputType) -> None:
        super().__init__(name, ruuid, key, type)


    def and_(self, x: bool) -> 'Cloud_Output_Bool':
        """And against x

        Returns a Cloud Output Bool on which further operations can be chained.
        """
        self._operations.append(
            (
                '**and', 
                tuple([x]), 
                {},
            )
        )

        return self


    def or_(self, x: bool) -> 'Cloud_Output_Bool':
        """Or against x

        Returns a Cloud Output Bool on which further operations can be chained.
        """
        self._operations.append(
            (
                '**or', 
                tuple([x]), 
                {},
            )
        )

        return self


    def xor_(self, x: bool) -> 'Cloud_Output_Bool':
        """Xor against x

        Returns a Cloud Output Bool on which further operations can be chained.
        """
        self._operations.append(
            (
                '**xor', 
                tuple([x]), 
                {},
            )
        )

        return self


    def not_(self) -> 'Cloud_Output_Bool':
        """Not Current Value

        Returns a Cloud Output Bool on which further operations can be chained.
        """
        # Special case because there is not underlying method to call for not so pass this hardcoded 
        # value and make sure that the operation interpreter respects this token as the 'not' operator.
        self._operations.append(
            (
                '**not', 
                (), 
                {},
            )
        )

        return self


class Cloud_Output_Int(Cloud_Output_Dynamic):
    """
    Cloud Output that will resolve to a Integer value after being retrieve or after all the operations have been executed
    """

    def __init__(self, name: str, ruuid: str, key: str, type: OutputType) -> None:
        super().__init__(name, ruuid, key, type)


    def add(self, x: int) -> 'Cloud_Output_Int':
        """Add x 

        Returns a Cloud Output Int on which further operations can be chained.
        """
        self._operations.append(
            (
                '__add__', 
                tuple([x]), 
                {},
            )
        )

        return self

    
    def subtract(self, x: int) -> 'Cloud_Output_Int':
        """Substract x

        Returns a Cloud Output Int on which further operations can be chained.
        """
        self._operations.append(
            (
                '__add__', 
                tuple([x*-1]), 
                {},
            )
        )

        return self


    def multiply(self, x: int) -> 'Cloud_Output_Int':
        """Multiply x

        Returns a Cloud Output Int on which further operations can be chained.
        """
        self._operations.append(
            (
                '__mul__', 
                tuple([x]), 
                {},
            )
        )

        return self


    def divide_mod(self, x: int) -> Tuple['Cloud_Output_Int', 'Cloud_Output_Int']:
        """Return the pair (i // x, i % x)

        Returns a Cloud Output Int on which further operations can be chained.
        """
        self._operations.append(
            (
                '__divmod__', 
                tuple([x]), 
                {},
            )
        )

        return self

    
class Cloud_Output_Str(Sequence, Cloud_Output_Dynamic):
    """
    Cloud Output that will resolve to a String value after being retrieve or after all the operations have been executed
    """


    def __init__(self, name: str, ruuid: str, key: str, type: OutputType) -> None:
        super().__init__(name, ruuid, key, type)

    def __len__(self):
        raise Exception

    def __getitem__(self, key) -> 'Cloud_Output_Str':
        self._operations.append(
            (
                '__getitem__',
                [key],
                {}
            )
        )    

        return self

    def __contains__(self, _o: str) -> Cloud_Output_Bool:
        self._operations.append(
            (
                '__contains__',
                [_o],
                {}
            )
        )

        return self

    def capitalize(self) -> 'Cloud_Output_Str':
        """Make the first character have upper case and the rest lower case.

        Append the operation to the Cloud Output Object and return the same Object for 
        any further operations.

        """
        self._operations.append(
            (
                'capitalize', 
                (), 
                {},
            )
        )
        return self

    def casefold(self) -> 'Cloud_Output_Str':
        """Make S suitable for caseless comparisons.

        Append the operation to the Cloud Output Object and return the same Object for 
        any further operations.
        
        """
        self._operations.append(
            (
                'casefold', 
                (), 
                {},
            )
        )
        return self

    def center(
        self, 
        width: SupportsIndex, 
        fillchar: str = None) -> 'Cloud_Output_Str':
        """S centered in a string of length width. Padding is done using the specified fill character (default is a space)

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """

        
        if fillchar:
            args = (width, fillchar)
        else:
            args = (width)


        self._operations.append(
            (
                'center', 
                args, 
                {},
            )
        )
        return self

    def count(self, 
        x: str, 
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Int:
        """The number of non-overlapping occurrences of substring sub in string S[start:end].
        
        Returns a Cloud Output Int on which further operations can be chained. The created Cloud Output Int
        will contain all previous operations of the current Cloud Output String that this is being called on.

        Optional arguments start and end are interpreted as in slice notation.
        """

        
        if start and end:
            args = (x, start, end)
        elif start and (not end):
            args = (x, start, None)

        elif (not start) and end:
            args = (x, None, end)

        else:
            args = (x, None, None)

        self._operations.append(
            (
                'count', 
                args, 
                {},
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def endswith(
        self, 
        suffix: Union[str, Tuple[str, ...]],
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Bool:
        """Return True if S ends with the specified suffix, False otherwise.
        
        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        
        With optional start, test S beginning at that position.
        With optional end, stop comparing S at that position.
        suffix can also be a tuple of strings to try.
        """

        args = []

        if suffix:
            args.append(suffix)

        if start:
            args.append(start)

        if end:
            args.append(end)


        self._operations.append(
            (
                'endswith', 
                tuple(args),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def expandtabs(self, tabsize=8) -> 'Cloud_Output_Str':
        """S where all tab characters are expanded using spaces.
        
        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        If tabsize is not given, a tab size of 8 characters is assumed.
        """
        self._operations.append(
            (
                'expandtabs', 
                (), 
                {
                    'tabsize': tabsize
                },
            )
        )
        return self

    def find(
        self, 
        sub: str, 
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Int:
        """
        Return the lowest index in S where substring sub is found, such that sub is contained within S[start:end].  
        
        Returns a Cloud Output Int on which further operations can be chained. The created Cloud Output Int
        will contain all previous operations of the current Cloud Output String that this is being called on.

        Optional arguments start and end are interpreted as in slice notation.
        """

        args = [sub]

        if start:
            args.append(start)

        if end:
            args.append(end)

        self._operations.append(
            (
                'find', 
                tuple(args), 
                {},
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def format(self, *args, **kwargs) -> 'Cloud_Output_Str':
        """Format S, using substitutions from args and kwargs.

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.

        The substitutions are identified by braces ('{' and '}').
        """
        self._operations.append(
            (
                'format', 
                args, 
                kwargs,
            )
        )
        return self

    def format_map(self, mapping) -> 'Cloud_Output_Str':
        """Format S, using substitutions from mapping.
        The substitutions are identified by braces ('{' and '}').

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """
        self._operations.append(
            (
                'format_map', 
                (), 
                {
                    'mapping': mapping
                },
            )
        )
        return self

    def index(
        self, 
        sub: str, 
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Int:
        """Return the lowest index in S where substring sub is found, such that sub is contained within S[start:end].  

        Returns a Cloud Output Int on which further operations can be chained. The created Cloud Output Int
        will contain all previous operations of the current Cloud Output String that this is being called on.
        
        Optional arguments start and end are interpreted as in slice notation.

        Raises ValueError when the substring is not found.
        """

        args = [sub]

        if start:
            args.append(start)

        if end:
            args.append(end)

        self._operations.append(
            (
                'index', 
                tuple(args), 
                {},
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isalnum(self) -> Cloud_Output_Bool:
        """Return True if all characters in S are alphanumeric and there is at least one character in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isalnum', 
                (), 
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv


    def isalpha(self) -> Cloud_Output_Bool:
        """Return True if all characters in S are alphabetic and there is at least one character in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isalpha',
                (),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isdecimal(self) -> Cloud_Output_Bool:
        """Return True if there are only decimal characters in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isdecimal', 
                (),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isdigit(self) -> Cloud_Output_Bool:
        """Return True if all characters in S are digits and there is at least one character in S, False otherwise.
        
        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isdigit',
                (),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isidentifier(self) -> Cloud_Output_Bool:
        """Return True if S is a valid identifier according to the language definition.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.

        Use keyword.iskeyword() to test for reserved identifiers
        such as "def" and "class".

        """
        self._operations.append(
            (
                'isidentifier',
                (),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def islower(self) -> Cloud_Output_Bool:
        """Return True if all cased characters in S are lowercase and there is at least one cased character in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'islower',
                (), 
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isnumeric(self) -> Cloud_Output_Bool:
        """Return True if there are only numeric characters in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isnumeric',
                (),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isprintable(self) -> Cloud_Output_Bool:
        """Return True if all characters in S are considered printable in repr() or S is empty, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isprintable', 
                (),
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isspace(self) -> Cloud_Output_Bool:
        """Return True if all characters in S are whitespace and there is at least one character in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isspace', 
                (), 
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def istitle(self) -> Cloud_Output_Bool:
        """Return True if S is a titlecased string and there is at least one
        character in S, i.e. upper- and titlecase characters may only
        follow uncased characters and lowercase characters only cased ones.
        Return False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'istitle', 
                (), 
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def isupper(self) -> Cloud_Output_Bool:
        """Return True if all cased characters in S are uppercase and there is
        at least one cased character in S, False otherwise.

        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        """
        self._operations.append(
            (
                'isupper', 
                (), 
                {},
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    

    def join(self, iterable: Iterable[str]) -> 'Cloud_Output_Str':
        """Return a string which is the concatenation of the strings in the iterable with S being the seperator.

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """
        self._operations.append(
            (
                'join',
                iterable, 
                {
                    
                },
            )
        )
        print(self._operations)
        return self


    def ljust(self, width: SupportsIndex, __fillchar: str = "") -> 'Cloud_Output_Str':
        """Return S left-justified in a Unicode string of length __width. 

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.

        Padding is done using the specified fill character (default is a space).
        """

        args = [width]

        if __fillchar:
            args.append(__fillchar)

        self._operations.append(
            (
                'ljust',
                tuple(args),
                {},
            )
        )
        return self

    def lower(self) -> 'Cloud_Output_Str':
        """Return a copy of the string S converted to lowercase.

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """
        self._operations.append(
            (
                'lower',
                (),
                {},
            )
        )
        return self

    def lstrip(self, chars: Optional[str] = None) -> 'Cloud_Output_Str':
        """Return a copy of the string S with leading whitespace removed.
        
        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        If chars is given and not None, remove characters in chars instead.
        """

        args = []

        if chars:
            args.append(chars)
        self._operations.append(
            (
                'lstrip', 
                tuple(args), 
                {},
            )
        )

        return self



    def replace(
        self, 
        old: str, 
        new: str, 
        count: SupportsIndex = None) -> 'Cloud_Output_Str':
        """Change all occurrences of substring old replaced by new. 

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        If the optional argument count is given, only the first count occurrences are replaced.
        """

        if count:
            args = (old, new, count)
        else:
            args = (old, new)

        self._operations.append(
            (
                'replace',
                args,
                {
                
                },
            )
        )

        return self

    def rfind(
        self, 
        sub: str, 
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Int:
        """Return the highest index in S where substring sub is found, such that sub is contained within S[start:end].  
        
        Returns a Cloud Output Int on which further operations can be chained. The created Cloud Output Int
        will contain all previous operations of the current Cloud Output String that this is being called on.

        Optional arguments start and end are interpreted as in slice notation.

        Return -1 on failure.
        """

        args = [sub]

        if start:
            args.append(start)

        if end:
            args.append(end)

        self._operations.append(
            (
                'rfind',
                tuple(args),
                {},
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def rindex(
        self, 
        sub: str,  
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Int:
        """Return the highest index in S where substring sub is found, such that sub is contained within S[start:end].  
        
        Returns a Cloud Output Int on which further operations can be chained. The created Cloud Output Int
        will contain all previous operations of the current Cloud Output String that this is being called on.

        Optional arguments start and end are interpreted as in slice notation.

        Raises ValueError when the substring is not found.
        """

        args = [sub]

        if start:
            args.append(start)

        if end:
            args.append(end)

        self._operations.append(
            (
                'rindex',
                (),
                {
                    "__sub": sub,
                    "__start": start,
                    "__end": end,   
                },
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv


    def rjust(
        self, 
        width: SupportsIndex, 
        fillchar: str = None) -> 'Cloud_Output_Str':
        """Right-justify S in a string of length width. Padding is
        done using the specified fill character (default is a space).

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """

        args = [width]

        if fillchar:
            args.append(fillchar)

        print(args)
        self._operations.append(
            (
                'rjust',
                tuple(args),
                {},
            )
        )

        return self


    #def rsplit(
    #    self, 
    #    sep: Optional[str] = None, 
    #    maxsplit: Optional[SupportsIndex] = None) -> List['Cloud_Output_Str']:
    #    """Return a list of the words in S, using sep as the
    #    delimiter string, starting at the end of the string and
    #    working to the front.  
#
    #    Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
    #    
    #    If maxsplit is given, at most maxsplit splits are done. 
    #    
    #    If sep is not specified, any whitespace string is a separator.
    #    """
    #    args = []
#
    #    if sep:
    #        args.append(sep)
#
    #    if maxsplit:
    #        args.append(maxsplit)
#
#
    #    self._operations.append(
    #        (
    #            'rsplit',
    #            tuple(args),
    #            {},
    #        )
    #    )
    #    return self

    
    def rstrip(
        self, 
        chars: Optional[str] = None)  -> 'Cloud_Output_Str':
        """Return S with trailing whitespace removed.
        
        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        If chars is given and not None, remove characters in chars instead.
        """
        self._operations.append(
            (
                'rstrip',
                (chars), 
                {},
            )
        )

    #def split(
    #    self, 
    #    sep=None,
    #    maxsplit=-1) -> List['Cloud_Output_Str']:
    #    """Return a list of the words in S, using sep as the delimiter string.  
#
    #    ## TODO Sequence type
    #    
    #    If maxsplit is given, at most maxsplit splits are done.
    #    
    #    If sep is not specified or is None, any whitespace string is a separator and empty strings are
    #    removed from the result.
    #    """
    #    self._operations.append(
    #        (
    #            'split', 
    #            (sep, maxsplit),
    #            {},
    #        )
    #    )
#
    #def splitlines(
    #    self, 
    #    keepends: bool = False) -> List['Cloud_Output_Str']:
    #    """
    #    Return a list of the lines in S, breaking at line boundaries.
#
    #    ## TODO Sequence type
    #    
    #    Line breaks are not included in the resulting list unless keepends is given and true.
    #    """
    #    self._operations.append(
    #        (
    #            'splitlines', 
    #            (keepends), 
    #            {},
    #        )
    #    )
#
    def startswith(
        self, 
        prefix: Union[str, Tuple[str, ...]], 
        start: Optional[SupportsIndex] = None, 
        end: Optional[SupportsIndex] = None) -> Cloud_Output_Int:
        """Return True if S starts with the specified prefix, False otherwise.
        
        Returns a Cloud Output Boolean on which further operations can be chained. The created Cloud Output Boolean
        will contain all previous operations of the current Cloud Output String that this is being called on.
        
        With optional start, test S beginning at that position.
        With optional end, stop comparing S at that position.
        prefix can also be a tuple of strings to try.
        """

        args = [prefix]

        if start:
            args.append(start)

        if end:
            args.append(end)

        self._operations.append(
            (
                'startswith', 
                tuple(args), 
                {},
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv

    def strip(
        self, 
        chars: Optional[str] = ...)  -> 'Cloud_Output_Str':
        """S with leading and trailing whitespace removed.
        
        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        If chars is given and not None, remove characters in chars instead.
        """
        self._operations.append(
            (
                'strip',
                (chars),
                {}
            )
        )
        return self

    def swapcase(self) -> 'Cloud_Output_Str':
        """S with uppercase characters converted to lowercase and vice versa.

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """
        self._operations.append(
            (
                'swapcase', 
                (), 
                {}
            )
        )

        return self


    def title(self) -> 'Cloud_Output_Str':
        """Titlecased version of S

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        i.e. words start with title case characters, all remaining cased characters have lower case.
        """
        self._operations.append(
            (
                'title',
                (),
                {}
            )
        )
        return self

    def translate(
        self,
        table: Mapping[
                int, 
                Union[ 
                    Union[int ,str,None], 
                    List[ Union[int, str, None] ] 
                ]
            ]
        ) -> 'Cloud_Output_Str':
        """S in which each character has been mapped through the given translation table. 
        
        The table must implement
        lookup/indexing via __getitem__, for instance a dictionary or list,
        mapping Unicode ordinals to Unicode ordinals, strings, or None. If
        this operation raises LookupError, the character is left untouched.
        Characters mapped to None are deleted.
        """
        self._operations.append(
            (
                'translate', 
                (table),
                {}
            )
        )

    def upper(self) -> 'Cloud_Output_Str':
        """
        Return a copy of S converted to uppercase.

        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        """
        self._operations.append(
            (
                'upper', 
                (), 
                {}
            )
        )

        return self


    def zfill(self, width: SupportsIndex) -> 'Cloud_Output_Str':
        """Pad a numeric string S with zeros on the left, to fill a field of the specified width. 
        
        Append the operation to the Cloud Output Object and return the same Cloud Output String object for any further operations.
        
        The string S is never truncated.
        """
        self._operations.append(
            (
                'zfill', 
                tuple([width]), 
                {}
            )
        )

        return self


# Wrapper type to denote all the possible single value outputs
T = TypeVar('T', Cloud_Output_Dynamic, Cloud_Output_Str, Cloud_Output_Int, Cloud_Output_Bool)

class Cloud_Output_Sequence(Sequence, Cloud_Output_Dynamic, Generic[T]):
    """
    Cloud Output that will resolve to a Sequence of values after being retrieve or after all the operations have been executed
    """
    def __init__(self, name: str, ruuid: str, key: str, type: OutputType, _member_class) -> None:
        super().__init__(name, ruuid, key, type)

        self._member_class = _member_class

    def __len__(self):
        raise Exception

    # Use these stub methods to define the typing signature so that the output is correct based on the input given
    @overload
    def __getitem__(self, key: slice) -> 'Cloud_Output_Sequence[T]': 
        pass

    @overload
    def  __getitem__(self, key: int) ->  T: 
        pass

    # implementation
    def __getitem__(self, key):
        self._operations.append(
                (
                    '__getitem__',
                    [key],
                    {}
                )
            )  
        
        if isinstance(key, slice):
            return self

        elif isinstance(key, int):
            rv = self._member_class(
                self._name, self._ruuid, self._key, self._type
            )

            rv._operations = self._operations.copy()

            return rv

        else:
            raise Exception


    def __contains__(self, o: Any) -> Cloud_Output_Bool:
        self._operations.append(
            (
                '__contains__',
                [o],
                {}
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv


    def __len__(self) -> Cloud_Output_Int:
        self._operations.append(
            (
                '__len__',
                [],
                {}
            )
        )

        rv = Cloud_Output_Int(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv



class Cloud_Output_Mapping(Sequence, Cloud_Output_Dynamic, Generic[T]):
    """
    Cloud Output that will resolve to a Mapping of string to values (after being retrieved or after all the operations have been executed).
    """
    def __init__(self, name: str, ruuid: str, key: str, type: OutputType, _member_class) -> None:
        super().__init__(name, ruuid, key, type)

        self._member_class = _member_class

    def __len__(self):
        raise Exception


    # implementation
    def __getitem__(self, key: str) -> T: 
        if not isinstance(key, str):
            raise Exception


        self._operations.append(
                (
                    '__getitem__',
                    [key],
                    {}
                )
            )  
        

        rv = self._member_class(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv


    def __contains__(self, _o: str) -> Cloud_Output_Bool:
        self._operations.append(
            (
                '__contains__',
                [_o],
                {}
            )
        )

        rv = Cloud_Output_Bool(
            self._name, self._ruuid, self._key, self._type
        )

        rv._operations = self._operations.copy()

        return rv



from rich import print
from typing import List

from core.constructs.output_manager import OutputManager

from ..constructs.project import Project



class CdevOutputManager(OutputManager):


    def print_header(self):
        myproject = Project.instance()

        print("")
        print(f"Project: {myproject.get_name()}")
        print("")


    def print_components_to_diff_against(self, old_component_names: List[str]):
        pass



from logging import raiseExceptions
import os
from pydantic import DirectoryPath
from pydantic.types import FilePath
from sortedcontainers.sortedlist import SortedKeyList
from typing import Dict, List, Tuple

from core.constructs.resource import (
    Resource,
    ResourceModel,
    ResourceReferenceModel,
    Resource_Reference,
)
from core.constructs.workspace import Workspace

from core.default.resources.simple.xlambda import (
    DependencyLayer,
    SimpleFunction,
    simple_function_model,
    SimpleFunctionConfiguration
)

from core.utils import hasher, module_loader, paths
from core.utils.logger import log

from serverless_parser import parser as serverless_parser


from . import writer
from . import package_mananger 


LAMBDA_LAYER_RUUID = "cdev::simple::lambda_layer"

def parse_folder(
    folder_path: DirectoryPath, 
) -> Tuple[List[ResourceModel], List[ResourceReferenceModel]]:
    """Search through the given folder looking for resource and references in Python files. 

    Args:
        folder_path (DirectoryPath): The directory to parse

    Returns:
        Tuple[
            List[ResourceModel],
            List[ResourceReferenceModel]
        ]
    
    Specifically, it loads all available python files and uses the loaded module to determine 
    the resources defined in the files. Any resource or reference defined in the global
    context of the file will be detected.
    
    Most resources are passed back as is, but there are optimizations performed on the `simple functions`. 
    Namely, Serverless functions are parsed to optimized the actual deployed artifact using the
    cparser library and then have their dependencies managed also.
    """
    if not os.path.isdir(folder_path):
        raise Exception

    log.debug("Finding resources in folder %s", folder_path)

    python_files = [
        f
        for f in os.listdir(folder_path)
        if os.path.isfile(os.path.join(folder_path, f)) and f[-3:] == ".py"
    ]

    # [{<resource>}]
    resources_rv = SortedKeyList(key=lambda x: x.hash)
    references_rv = SortedKeyList(key=lambda x: x.hash)


    for pf in python_files:
        found_resources, found_references = _find_resources_information_from_file(
            os.path.join(folder_path, pf)
        )

        if found_resources:
            resources_rv.update(found_resources)

        if found_references:
            references_rv.update(found_references)

    # Any duplicate layers can be removed 
    cleaned_resources_rv = _deduplicate_resources_list(resources_rv)

    return cleaned_resources_rv, references_rv


def _deduplicate_resources_list(resources: List[Resource]) -> List[Resource]:
    """Remove duplicated layer resources

    Args:
        resources (List[Resource]): Sorted List of resources by x.hash

    Returns:
        resource (List[Resources]): List with duplicate resources removed

    Since multiple functions can produce the same Layer resource by referencing the same 
    3rd party resource, we need to deduplicate the layers from the list.
    """

    for i in range(len(resources)):
        if i+1 >= len(resources):
            break

        if (resources[i].hash == resources[i+1].hash) and (resources[i].ruuid == resources[i+1].ruuid) and (resources[i].name == resources[i+1].name):
            resources.pop(i)


    return resources

def _get_module_name_from_path(fp: FilePath):
    """Convert a full file path of a python path into a importable module name

    Args:
        fp (FilePath): path to file

    Returns:
        str: The importable python module name 


    All module names will end up being relative to the workspace path. Note that this means
    the `Workspace` base path should be on the `Python Path`. This usually happens by default
    because the `Workspace` starts from the cwd. 
    """
    relative_to_project_path = paths.get_relative_to_workspace_path(fp)

    relative_to_project_path_parts = relative_to_project_path.split("/")

    # If the last part of the file is __init__.pt then python will import it when the 
    # rest of the path is given without the last part
    if relative_to_project_path_parts[-1] == "__init__.py":
        relative_to_project_path_parts.pop()
    else:
        # remove the .py part of the file name
        relative_to_project_path_parts[-1] = relative_to_project_path_parts[-1][:-3]

    # join the parts back with '.' to create the valid python module name
    full_module_path_from_project = ".".join(relative_to_project_path_parts)

    return full_module_path_from_project


def _find_resources_information_from_file(
    fp: FilePath,
) -> Tuple[List[ResourceModel], List[ResourceReferenceModel]]:
    """Load a file and find top level objects that are Resources or References

    Args:
        fp (FilePath): path to python file

    Raises:
        Exception: [description]

    Returns:
        Tuple[List[ResourceModel], List[ResourceReferenceModel]]: Resources and References
    """
    # Input: filepath
    if not os.path.isfile(fp):
        raise Exception

    if not fp[-3:] == '.py':
        raise Exception

    mod_name = _get_module_name_from_path(fp)

    # When the python file is imported and executed all the Cdev resources are created
    mod = module_loader.import_module(mod_name, denote_output=True)

    resource_rv = []
    reference_rv = []

    functions_to_parse: List[str] = []
    function_name_to_info: Dict[str, simple_function_model] = {}


    for i in dir(mod):
        obj = getattr(mod, i)

        if isinstance(obj, Resource):
            # Find all the Resources in the module and render them

            if isinstance(obj, SimpleFunction):
                # Functions are a special case as they will go through the parser and the output
                # of that will be the returned resource
                functions_to_parse.append(obj.configuration.handler)
                function_name_to_info[obj.configuration.handler] = obj
                
            else:
                resource_rv.append(obj.render())

        elif isinstance(obj, Resource_Reference):
            reference_rv.append(obj.render())

    if functions_to_parse:
        log.debug("Parsing functions (%s) from %s", functions_to_parse, fp)
        parsed_function_info, parsed_dependency_info = _parse_serverless_functions(
            fp, 
            functions_to_parse,
            handler_name_to_info=function_name_to_info
        )

        resource_rv.extend(parsed_function_info)
        resource_rv.extend(parsed_dependency_info)
        

    return resource_rv, reference_rv



def _parse_serverless_functions(
    filepath: FilePath,
    functions_names_to_parse: List[str],
    handler_name_to_info: Dict[str, SimpleFunction],
    manual_includes: Dict = {},
    global_includes: List = [],
) -> Tuple[List[simple_function_model], List[DependencyLayer]]:
    """Parse a given set of function names from a given file

    Args:
        filepath (FilePath): The original file
        functions_names_to_parse (List[str]): functions to parse
        handler_name_to_info (Dict[str, SimpleFunction]): dict of additional information 
        manual_includes (Dict, optional): Dict of information about extra lines to include. Defaults to {}.
        global_includes (List, optional): List of global lines to include. Defaults to [].

    Returns:
        Tuple[List[simple_function_model], List[DependencyLayer]]: Functions and Dependencies parsed

    Use the `serverless_parser` library to get information about each desired function and its dependencies.
    Then use that information to create the needed archives for the functions and return the information as
    Resources.
    """

    # Get all the info about a set of functions from the original file
    parsed_file_info = serverless_parser.parse_functions_from_file(
        filepath, include_functions=functions_names_to_parse, remove_top_annotation=True
    )

    # Base path that the all the archives will go 
    base_archive_path = os.path.join(
        Workspace.instance().settings.INTERMEDIATE_FOLDER_LOCATION,
        Workspace.instance().get_resource_state_uuid(),
    )

    if not os.path.isdir(base_archive_path):
        os.mkdir(base_archive_path)

    if Workspace.instance().settings.USE_DOCKER:
        # IF the user has denoted that they want to use Docker to build dependencies for 
        # other architectures, then make sure a downloads cache is set up
        download_cache = os.path.join(
            Workspace.instance().settings.INTERMEDIATE_FOLDER_LOCATION,
            ".download_cache",
        )

        if not os.path.isdir(download_cache):
            os.mkdir(download_cache)

    else:
        download_cache = None
    
    rv = []
    seen_layers = {}

    for parsed_function in parsed_file_info.parsed_functions:

        previous_info = handler_name_to_info.get(parsed_function.name)
        needed_module_information = package_mananger.get_top_level_module_info(
            parsed_function.imported_packages, filepath, previous_info.platform, download_cache
        )

        log.debug("Needed modules (%s) for %s", needed_module_information, parsed_function.name)

        (
            handler_archive_path,
            handler_archive_hash,
            base_handler_path,
            dependencies_info,
        ) = writer.create_full_deployment_package(
            filepath,
            parsed_function.get_line_numbers_serializeable(),
            parsed_function.name,
            base_archive_path,
            pkgs=needed_module_information,
        )
        
        
        # Update the seen packages 
        if dependencies_info:
            # Helps return only one copy of each layer for the file
            # change the path of the artifact to a relative path to the workspace
            for dependency in dependencies_info:
                dependency.artifact_path = paths.get_relative_to_workspace_path(dependency.artifact_path)

            seen_layers.update(
              {x.hash:x.render() for x in dependencies_info}
            )

        handler_path = base_handler_path + "." + parsed_function.name

        new_configuration = SimpleFunctionConfiguration(
            handler=handler_path,
            description=previous_info.configuration.description,
            environment_variables=previous_info.configuration.environment_variables
        )

        new_function = SimpleFunction(
            cdev_name=previous_info.name,
            filepath=paths.get_relative_to_workspace_path(handler_archive_path),
            events=previous_info.events,
            configuration=new_configuration,
            function_permissions=previous_info.granted_permissions,
            external_dependencies=dependencies_info if dependencies_info else [],
            src_code_hash=handler_archive_hash,
            nonce=previous_info.nonce,
            platform=previous_info.platform
        )

    
        rv.append(new_function.render())

    return rv, [layer for _,layer in seen_layers.items()]



from typing import Any, Dict, List
from uuid import uuid4

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.default.resources.simple import topic 
from core.constructs.output_manager import OutputTask
from core.utils import hasher

from .. import aws_client

def _create_simple_topic(
    transaction_token: str, 
    namespace_token: str, 
    resource:  topic.simple_topic_model,
    output_task: OutputTask
) -> Dict:

    attributes = {}

    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])

    topic_name = f"cdev-topic-{full_namespace_suffix}.fifo" if resource.is_fifo else f"cdev-topic-{full_namespace_suffix}"

    if resource.is_fifo:
        attributes.update({"FifoTopic": str(resource.is_fifo)})

    output_task.update(comment=f'Creating Topic {topic_name}')

    rv = aws_client.run_client_function(
        "sns", "create_topic", {"Name": topic_name, "Attributes": attributes}
    )
    
    output_info = {
        "topic_name": topic_name,
        "cdev_name": resource.name,
        "arn": rv.get("TopicArn"),
        "cloud_id": rv.get("TopicArn"),
    }

   
    return output_info


def _update_simple_topic(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: topic.simple_topic_model,
    new_resource: topic.simple_topic_model,
    previous_output: Dict,
    output_task: OutputTask
) -> Dict:
    _remove_simple_topic(transaction_token, previous_output, output_task)
    new_rv = _create_simple_topic(transaction_token, namespace_token, new_resource, output_task)
    return new_rv


def _remove_simple_topic(
    transaction_token: str, previous_output: Dict, output_task: OutputTask
):
    topic_arn = previous_output.get("cloud_id")

    aws_client.run_client_function("sns", "delete_topic", {"TopicArn": topic_arn})


def add_subscriber(topic_arn: str, protocol: str, endpoint: str) -> str:
    rv = aws_client.run_client_function(
        "sns",
        "subscribe",
        {
            "TopicArn": topic_arn,
            "Protocol": protocol,
            "Endpoint": endpoint,
            "ReturnSubscriptionArn": True,
        },
    )

    return rv.get("SubscriptionArn")


def remove_subscriber(subscription_arn: str):
    rv = aws_client.run_client_function(
        "sns", "unsubscribe", {"SubscriptionArn": subscription_arn}
    )


def handle_simple_topic_deployment(
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_topic(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
        return _update_simple_topic(
            transaction_token,
            namespace_token,
            topic.simple_topic_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )

    elif resource_diff.action_type == Resource_Change_Type.DELETE:
        _remove_simple_topic(
            transaction_token,
            previous_output,
            output_task
        )

        return {}



"""Utilities for displaying data in a unified and formatted manner



"""

from rich.console import Console
from typing import Any, List, Tuple, Union

from rich.progress import Progress, TaskID  

from core.constructs.components import Component_Change_Type, ComponentModel, Component_Difference
from core.constructs.resource import Resource_Change_Type, Resource_Difference, Resource_Reference_Change_Type, Resource_Reference_Difference

class OutputManager():
    def __init__(self, console: Console=None, progress: Progress=None) -> None:

        if console:
            self._console = console
        else:
            self._console = Console()

        if progress:
            self._progress = progress
        else:
            self._progress = None


    def print_header(self, resource_state_uuid: str):        
        print(f"Resource State: {resource_state_uuid}")

        print("")


    def print_local_state(self, rendered_components: List[ComponentModel]):

        rendered_components.sort(key=lambda x: x.name)
        
        self._console.print(f"Current State:")
        for component in rendered_components:
            self._console.print(f"Component: [bold blue]{component.name}[/bold blue]")

            if component.resources:
                self._console.print(f"    Resources:")
                for resource in component.resources:
                    self._console.print(f"        [bold blue]{resource.name} ({resource.ruuid})[/bold blue]")


            if component.references:
                self._console.print(f"    References:")
                for reference in component.references:
                    self._console.print(f"        [bold blue]From {reference.component_name} reference {reference.name} ({reference.ruuid})[/bold blue]")

        self._console.print("")


    def print_components_to_diff_against(self, old_component_names: List[str]):

        self._console.print("Components to diff against:")
        for component_name in old_component_names:
            self._console.print(f"    {component_name}")


        self._console.print("")


    def print_state_differences(self, differences: Tuple[List[Component_Difference], List[Resource_Difference], List[Resource_Reference_Difference]]):
        component_differences = differences[0]
        resource_differences = differences[1]
        reference_differences = differences[2]


        self._console.print(f"[bold white]Differences in State:[/bold white]")

        for component_diff in component_differences:
            if component_diff.action_type == Component_Change_Type.UPDATE_NAME:
                self._console.print(f"    [bold yellow]Update Name: [/bold yellow][bold blue]{component_diff.previous_name} to {component_diff.new_name} (component)[/bold blue]")
                continue

            elif component_diff.action_type == Component_Change_Type.UPDATE_IDENTITY:
                self._console.print(f"    [bold yellow]Update Identity: [/bold yellow][bold blue]{component_diff.new_name} (component)[/bold blue]")

            elif component_diff.action_type == Component_Change_Type.CREATE:
                self._console.print(f"    [bold green]Create: [/bold green][bold blue]{component_diff.new_name} (component)[/bold blue]")

            elif component_diff.action_type == Component_Change_Type.DELETE:
                self._console.print(f"    [bold red]Delete: [/bold red] [bold blue]{component_diff.previous_name} (component)[/bold blue]")


            resource_changes = [x for x in resource_differences if x.component_name == component_diff.new_name or x.component_name == component_diff.previous_name ]

            if resource_changes:
                for resource_diff in resource_changes:
                    if resource_diff.action_type == Resource_Change_Type.CREATE:
                        self._console.print(f"        [bold green]Create:[/bold green][bold blue] {resource_diff.new_resource.name} ({resource_diff.new_resource.ruuid})[/bold blue]")

                    elif resource_diff.action_type == Resource_Change_Type.DELETE:
                        self._console.print(f"        [bold red]Delete:[/bold red][bold blue] {resource_diff.previous_resource.name} ({resource_diff.previous_resource.ruuid})[/bold blue]")

                    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
                        self._console.print(f"        [bold yellow]Update:[/bold yellow][bold blue] {resource_diff.new_resource.name} ({resource_diff.new_resource.ruuid})[/bold blue]")

                    elif resource_diff.action_type == Resource_Change_Type.UPDATE_NAME:
                        self._console.print(f"        [bold yellow]Update Name:[/bold yellow][bold blue] from {resource_diff.previous_resource.name} to {resource_diff.new_resource.name} ({resource_diff.new_resource.ruuid})[/bold blue]")


            reference_changes = [x for x in reference_differences if x.originating_component_name == component_diff.new_name]

            if reference_changes:
                for reference_diff in reference_changes:
                    if reference_diff.action_type == Resource_Reference_Change_Type.CREATE:
                        self._console.print(f"        [bold green]Create reference:[/bold green][bold blue] {reference_diff.resource_reference.name} ({reference_diff.resource_reference.ruuid}) from {reference_diff.originating_component_name}[/bold blue]")

                    elif reference_diff.action_type == Resource_Reference_Change_Type.DELETE:
                        self._console.print(f"        [bold red]Delete reference:[/bold red][bold blue] {reference_diff.resource_reference.name} ({reference_diff.resource_reference.ruuid}) from {reference_diff.originating_component_name}[/bold blue]")


    def create_task(self, description: str, start: bool = True, total: int = 100, completed: int = 0, visible: bool = True, **fields: Any) -> 'OutputTask':
        """Create a task for the progress object in this output manager. 

        Raises:
            Exception: [description]

        Returns:
            OutputTask: Object to use to track the progress of the event
        """


        if not self._progress:
            raise Exception


        task_id = self._progress.add_task(description=description, start=start, total=total, completed=completed, visible=visible, **fields)

        return OutputTask(self, task_id)


    def create_output_description(self, node: Union[Resource_Difference, Resource_Reference_Difference, Component_Difference]) -> str:
        if isinstance(node, Resource_Difference):
            if node.action_type == Resource_Change_Type.CREATE:
                return f"[bold green]Creating:[/bold green][bold blue] {node.new_resource.name} ({node.new_resource.ruuid})[/bold blue]"

            elif node.action_type == Resource_Change_Type.DELETE:
                return f"[bold red]Deleting:[/bold red][bold blue] {node.previous_resource.name} ({node.previous_resource.ruuid})[/bold blue]"

            elif node.action_type == Resource_Change_Type.UPDATE_IDENTITY:
                return f"[bold yellow]Updating:[/bold yellow][bold blue] {node.new_resource.name} ({node.new_resource.ruuid})[/bold blue]"

            elif node.action_type == Resource_Change_Type.UPDATE_NAME:
                return f"[bold yellow]Updating Name:[/bold yellow][bold blue] from {node.previous_resource.name} to {node.new_resource.name} ({node.new_resource.ruuid})[/bold blue]"

        elif isinstance(node, Resource_Reference_Difference):
            if node.action_type == Resource_Reference_Change_Type.CREATE:
                return f"[bold green]Create reference:[/bold green][bold blue] {node.resource_reference.name} ({node.resource_reference.ruuid}) from {node.originating_component_name}[/bold blue]"

            elif node.action_type == Resource_Reference_Change_Type.DELETE:
                return f"[bold red]Delete reference:[/bold red][bold blue] {node.resource_reference.name} ({node.resource_reference.ruuid}) from {node.originating_component_name}[/bold blue]"

        elif isinstance(node, Component_Difference):
            if node.action_type == Component_Change_Type.UPDATE_NAME:
                return f"[bold yellow]Update Name: [/bold yellow][bold blue]{node.previous_name} to {node.new_name} (component)[/bold blue]"

            elif node.action_type == Component_Change_Type.UPDATE_IDENTITY:
                return f"[bold yellow]Update Identity: [/bold yellow][bold blue]{node.new_name} (component)[/bold blue]"

            elif node.action_type == Component_Change_Type.CREATE:
                return f"[bold green]Create: [/bold green][bold blue]{node.new_name} (component)[/bold blue]"

            elif node.action_type == Component_Change_Type.DELETE:
                return f"[bold red]Delete: [/bold red] [bold blue]{node.previous_name} (component)[/bold blue]"

        else:
            raise Exception(f"Trying to deploy node {node} but it is not a correct type ")


class OutputTask():
    """
    Wrapper around an output that can be used to track long running events. The implementation is as a wrapper around a 'progress task' from the
    rich library. Thus, the api for 'update' and 'start' follows the same pattern as that of a 'task' in the rich library. The print method provides
    access for sending messages to the parent console of this task. 
    """  

    def __init__(self, output_manager: OutputManager, task_id: TaskID) -> None:
        self._output_manager = output_manager
        self._task_id = task_id

    
    def print(self, msg: str):
        """Print a message to the parent output context.

        Args:
            msg (str): message to print
        """
        self._output_manager._console.print(msg)


    def print_error(self, msg: str):
        """
        Print a message as an error

        Args:
            msg (str): error message
        """
        self.print(f"[bold red]{msg}[/bold red]")



    def update(self, *args, total: float = None, completed: float = None, advance: None = None, description: str = None, visible: bool = None, refresh: bool = False, **fields: Any):
        #self._output_manager._console.print(fields.get('comment'))
        self._output_manager._progress.update(self._task_id, *args, total=total, completed=completed, advance=advance, description=description, visible=visible, refresh=refresh, **fields)


    def start_task(self):
        self._output_manager._progress.start_task(self._task_id)



"""Set of utilities for working with immutable data within the framework
"""
from collections.abc import Mapping
from enum import Enum
from pydantic import BaseModel

from pydantic.utils import (
    ROOT_KEY,
    ValueItems,
    sequence_like,
)

from pydantic.typing import (
    is_namedtuple,
)

from typing import  Any

class frozendict(Mapping):

    def __init__(self, d: dict):
        self._d = d

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        return hash(tuple(sorted(self._d.items())))

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        
        if not (isinstance(v, frozendict)) :
            
            raise TypeError(f'{v} Not Frozen Dict')

        return v
    
    def __repr__(self) -> str:
        return f"FrozenDict({self._d})"


class ImmutableModel(BaseModel):
    @classmethod
    def _get_value(
        cls,
        v: Any,
        to_dict: bool,
        by_alias: bool,
        include ,
        exclude,
        exclude_unset: bool,
        exclude_defaults: bool,
        exclude_none: bool,
    ) -> Any:

        if isinstance(v, BaseModel):
            if to_dict:
                v_dict = frozendict( v.dict(
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    include=include,
                    exclude=exclude,
                    exclude_none=exclude_none,
                ))
                if ROOT_KEY in v_dict:
                    return v_dict[ROOT_KEY]
                return v_dict
            else:
                return v.copy(include=include, exclude=exclude)

        value_exclude = ValueItems(v, exclude) if exclude else None
        value_include = ValueItems(v, include) if include else None

        if isinstance(v, dict):
            return frozendict( {
                k_: cls._get_value(
                    v_,
                    to_dict=to_dict,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    include=value_include and value_include.for_element(k_),
                    exclude=value_exclude and value_exclude.for_element(k_),
                    exclude_none=exclude_none,
                )
                for k_, v_ in v.items()
                if (not value_exclude or not value_exclude.is_excluded(k_))
                and (not value_include or value_include.is_included(k_))
            })

        elif sequence_like(v):
            seq_args = (
                cls._get_value(
                    v_,
                    to_dict=to_dict,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    include=value_include and value_include.for_element(i),
                    exclude=value_exclude and value_exclude.for_element(i),
                    exclude_none=exclude_none,
                )
                for i, v_ in enumerate(v)
                if (not value_exclude or not value_exclude.is_excluded(i))
                and (not value_include or value_include.is_included(i))
            )
            
            return v.__class__(*seq_args) if is_namedtuple(v.__class__) else v.__class__(seq_args)

        elif isinstance(v, Enum) and getattr(cls.Config, 'use_enum_values', False):
            return v.value

        else:
            return v

    class Config:
        extra = "allow"
        frozen = True


        



## This module contains the code to parse functions out of python files. This will be used to parse the function handlers into individual files
## to help deploy them as individual primitives. This module will hold no state and instead simply parse what is available in a file. It will be
## dependant on other modules to monitor the state of files to determine when an update has occured. This module only provides information about
## what lines correspond to a function and does not handle the actual writing of intermediate files.
